# Copyright (C) 2005-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

__docformat__ = "google"

from typing import TYPE_CHECKING, Iterable, List, Optional, Type

from .lazy_import import lazy_import

lazy_import(globals(), """
import time

from breezy import (
    config,
    osutils,
    )
from breezy.i18n import gettext
""")

from . import controldir, debug, errors, graph, registry
from . import revision as _mod_revision
from . import ui
from .decorators import only_raises
from .inter import InterObject
from .lock import LogicalLockResult, _RelockDebugMixin
from .revisiontree import RevisionTree
from .trace import (log_exception_quietly, mutter, mutter_callsite, note,
                    warning)

# Old formats display a warning, but only once
_deprecation_warning_done = False


class IsInWriteGroupError(errors.InternalBzrError):

    _fmt = "May not refresh_data of repo %(repo)s while in a write group."

    def __init__(self, repo):
        errors.InternalBzrError.__init__(self, repo=repo)


class CannotSetRevisionId(errors.BzrError):

    _fmt = "Repository format does not support setting revision ids."


class FetchResult:
    """Result of a fetch operation.

    Attributes:
      revidmap: For lossy fetches, map from source revid to target revid.
      total_fetched: Number of revisions fetched
    """

    def __init__(self, total_fetched=None, revidmap=None):
        self.total_fetched = total_fetched
        self.revidmap = revidmap


class CommitBuilder:
    """Provides an interface to build up a commit.

    This allows describing a tree to be committed without needing to
    know the internals of the format of the repository.
    """

    # all clients should supply tree roots.
    record_root_entry = True
    # whether this commit builder will automatically update the branch that is
    # being committed to
    updates_branch = False

    def __init__(self, repository, parents, config_stack, timestamp=None,
                 timezone=None, committer=None, revprops=None,
                 revision_id=None, lossy=False):
        """Initiate a CommitBuilder.

        Args:
          repository: Repository to commit to.
          parents: Revision ids of the parents of the new revision.
          timestamp: Optional timestamp recorded for commit.
          timezone: Optional timezone for timestamp.
          committer: Optional committer to set for commit.
          revprops: Optional dictionary of revision properties.
          revision_id: Optional revision id.
          lossy: Whether to discard data that can not be natively
            represented, when pushing to a foreign VCS
        """
        self._config_stack = config_stack
        self._lossy = lossy

        if committer is None:
            self._committer = self._config_stack.get('email')
        elif not isinstance(committer, str):
            self._committer = committer.decode()  # throw if non-ascii
        else:
            self._committer = committer

        self.parents = parents
        self.repository = repository

        self._revprops = {}
        if revprops is not None:
            self._validate_revprops(revprops)
            self._revprops.update(revprops)

        if timestamp is None:
            timestamp = time.time()
        # Restrict resolution to 1ms
        self._timestamp = round(timestamp, 3)

        if timezone is None:
            self._timezone = osutils.local_time_offset()
        else:
            self._timezone = int(timezone)

        self._generate_revision_if_needed(revision_id)

    def any_changes(self):
        """Return True if any entries were changed.

        This includes merge-only changes. It is the core for the --unchanged
        detection in commit.

        Returns: True if any changes have occured.
        """
        raise NotImplementedError(self.any_changes)

    def _validate_unicode_text(self, text, context):
        """Verify things like commit messages don't have bogus characters."""
        # TODO(jelmer): Make this repository-format specific
        if '\r' in text:
            raise ValueError('Invalid value for {}: {!r}'.format(context, text))

    def _validate_revprops(self, revprops):
        for key, value in revprops.items():
            # We know that the XML serializers do not round trip '\r'
            # correctly, so refuse to accept them
            if not isinstance(value, str):
                raise ValueError('revision property (%s) is not a valid'
                                 ' (unicode) string: %r' % (key, value))
            # TODO(jelmer): Make this repository-format specific
            self._validate_unicode_text(value,
                                        'revision property ({})'.format(key))

    def commit(self, message):
        """Make the actual commit.

        Returns: The revision id of the recorded revision.
        """
        raise NotImplementedError(self.commit)

    def abort(self):
        """Abort the commit that is being built.
        """
        raise NotImplementedError(self.abort)

    def revision_tree(self) -> "RevisionTree":
        """Return the tree that was just committed.

        After calling commit() this can be called to get a
        RevisionTree representing the newly committed tree. This is
        preferred to calling Repository.revision_tree() because that may
        require deserializing the inventory, while we already have a copy in
        memory.
        """
        raise NotImplementedError(self.revision_tree)

    def finish_inventory(self):
        """Tell the builder that the inventory is finished.

        Returns: The inventory id in the repository, which can be used with
            repository.get_inventory.
        """
        raise NotImplementedError(self.finish_inventory)

    def _generate_revision_if_needed(self, revision_id):
        """Create a revision id if None was supplied.

        If the repository can not support user-specified revision ids
        they should override this function and raise CannotSetRevisionId
        if _new_revision_id is not None.

        Raises:
          CannotSetRevisionId
        """
        if not self.repository._format.supports_setting_revision_ids:
            if revision_id is not None:
                raise CannotSetRevisionId()
            return
        if revision_id is None:
            self._new_revision_id = self._gen_revision_id()
            self.random_revid = True
        else:
            self._new_revision_id = revision_id
            self.random_revid = False

    def record_iter_changes(self, tree, basis_revision_id, iter_changes):
        """Record a new tree via iter_changes.

        Args:
          tree: The tree to obtain text contents from for changed objects.
          basis_revision_id: The revision id of the tree the iter_changes
            has been generated against. Currently assumed to be the same
            as self.parents[0] - if it is not, errors may occur.
          iter_changes: An iter_changes iterator with the changes to apply
            to basis_revision_id. The iterator must not include any items with
            a current kind of None - missing items must be either filtered out
            or errored-on beefore record_iter_changes sees the item.
        Returns: A generator of (relpath, fs_hash) tuples for use with
            tree._observed_sha1.
        """
        raise NotImplementedError(self.record_iter_changes)


class RepositoryWriteLockResult(LogicalLockResult):
    """The result of write locking a repository.

    Attributes:
      repository_token: The token obtained from the underlying lock, or
        None.
      unlock: A callable which will unlock the lock.
    """

    def __init__(self, unlock, repository_token):
        LogicalLockResult.__init__(self, unlock)
        self.repository_token = repository_token

    def __repr__(self):
        return "RepositoryWriteLockResult({}, {})".format(self.repository_token,
                                                      self.unlock)


class WriteGroup:
    """Context manager that manages a write group.

    Raising an exception will result in the write group being aborted.
    """

    def __init__(self, repository, suppress_errors=False):
        self.repository = repository
        self._suppress_errors = suppress_errors

    def __enter__(self):
        self.repository.start_write_group()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            self.repository.abort_write_group(self._suppress_errors)
            return False
        else:
            self.repository.commit_write_group()


######################################################################
# Repositories


class Repository(controldir.ControlComponent, _RelockDebugMixin):
    """Repository holding history for one or more branches.

    The repository holds and retrieves historical information including
    revisions and file history.  It's normally accessed only by the Branch,
    which views a particular line of development through that history.

    See VersionedFileRepository in breezy.vf_repository for the
    base class for most Bazaar repositories.
    """

    # Does this repository implementation support random access to
    # items in the tree, or just bulk fetching/pushing of data?
    supports_random_access = True

    def abort_write_group(self, suppress_errors=False):
        """Commit the contents accrued within the current write group.

        Args:
          suppress_errors: if true, abort_write_group will catch and log
            unexpected errors that happen during the abort, rather than
            allowing them to propagate.  Defaults to False.
        """
        if self._write_group is not self.get_transaction():
            # has an unlock or relock occured ?
            if suppress_errors:
                mutter(
                    '(suppressed) mismatched lock context and write group. %r, %r',
                    self._write_group, self.get_transaction())
                return
            raise errors.BzrError(
                'mismatched lock context and write group. %r, %r' %
                (self._write_group, self.get_transaction()))
        try:
            self._abort_write_group()
        except Exception as exc:
            self._write_group = None
            if not suppress_errors:
                raise
            mutter('abort_write_group failed')
            log_exception_quietly()
            note(gettext('brz: ERROR (ignored): %s'), exc)
        self._write_group = None

    def _abort_write_group(self):
        """Template method for per-repository write group cleanup.

        This is called during abort before the write group is considered to be
        finished and should cleanup any internal state accrued during the write
        group. There is no requirement that data handed to the repository be
        *not* made available - this is not a rollback - but neither should any
        attempt be made to ensure that data added is fully commited. Abort is
        invoked when an error has occured so futher disk or network operations
        may not be possible or may error and if possible should not be
        attempted.
        """

    def add_fallback_repository(self, repository):
        """Add a repository to use for looking up data not held locally.

        Args:
          repository: A repository.
        """
        raise NotImplementedError(self.add_fallback_repository)

    def _check_fallback_repository(self, repository):
        """Check that this repository can fallback to repository safely.

        Raise an error if not.

        Args:
          repository: A repository to fallback to.
        """
        return InterRepository._assert_same_model(self, repository)

    def all_revision_ids(self):
        """Returns a list of all the revision ids in the repository.

        This is conceptually deprecated because code should generally work on
        the graph reachable from a particular revision, and ignore any other
        revisions that might be present.  There is no direct replacement
        method.
        """
        if 'evil' in debug.debug_flags:
            mutter_callsite(2, "all_revision_ids is linear with history.")
        return self._all_revision_ids()

    def _all_revision_ids(self):
        """Returns a list of all the revision ids in the repository.

        These are in as much topological order as the underlying store can
        present.
        """
        raise NotImplementedError(self._all_revision_ids)

    def break_lock(self):
        """Break a lock if one is present from another instance.

        Uses the ui factory to ask for confirmation if the lock may be from
        an active process.
        """
        self.control_files.break_lock()

    @staticmethod
    def create(controldir):
        """Construct the current default format repository in controldir."""
        return RepositoryFormat.get_default_format().initialize(controldir)

    def __init__(self, _format, controldir, control_files):
        """instantiate a Repository.

        Args:
          _format: The format of the repository on disk.
          controldir: The ControlDir of the repository.
          control_files: Control files to use for locking, etc.
        """
        # In the future we will have a single api for all stores for
        # getting file texts, inventories and revisions, then
        # this construct will accept instances of those things.
        super().__init__()
        self._format = _format
        # the following are part of the public API for Repository:
        self.controldir = controldir
        self.control_files = control_files
        # for tests
        self._write_group = None
        # Additional places to query for data.
        self._fallback_repositories = []

    @property
    def user_transport(self):
        return self.controldir.user_transport

    @property
    def control_transport(self):
        return self._transport

    def __repr__(self):
        if self._fallback_repositories:
            return '{}({!r}, fallback_repositories={!r})'.format(
                self.__class__.__name__,
                self.base,
                self._fallback_repositories)
        else:
            return '{}({!r})'.format(self.__class__.__name__,
                               self.base)

    def _has_same_fallbacks(self, other_repo):
        """Returns true if the repositories have the same fallbacks."""
        my_fb = self._fallback_repositories
        other_fb = other_repo._fallback_repositories
        if len(my_fb) != len(other_fb):
            return False
        for f, g in zip(my_fb, other_fb):
            if not f.has_same_location(g):
                return False
        return True

    def has_same_location(self, other):
        """Returns a boolean indicating if this repository is at the same
        location as another repository.

        This might return False even when two repository objects are accessing
        the same physical repository via different URLs.
        """
        if self.__class__ is not other.__class__:
            return False
        return (self.control_url == other.control_url)

    def is_in_write_group(self):
        """Return True if there is an open write group.

        :seealso: start_write_group.
        """
        return self._write_group is not None

    def is_locked(self):
        return self.control_files.is_locked()

    def is_write_locked(self):
        """Return True if this object is write locked."""
        return self.is_locked() and self.control_files._lock_mode == 'w'

    def lock_write(self, token=None):
        """Lock this repository for writing.

        This causes caching within the repository obejct to start accumlating
        data during reads, and allows a 'write_group' to be obtained. Write
        groups must be used for actual data insertion.

        A token should be passed in if you know that you have locked the object
        some other way, and need to synchronise this object's state with that
        fact.

        XXX: this docstring is duplicated in many places, e.g. lockable_files.py

        Args:
          token: if this is already locked, then lock_write will fail
            unless the token matches the existing lock.
        Returns:
          a token if this instance supports tokens, otherwise None.

        Raises:
          TokenLockingNotSupported: when a token is given but this
            instance doesn't support using token locks.
          MismatchedToken: if the specified token doesn't match the token
            of the existing lock.

        Returns:
          A RepositoryWriteLockResult.
        """
        locked = self.is_locked()
        token = self.control_files.lock_write(token=token)
        if not locked:
            self._warn_if_deprecated()
            self._note_lock('w')
            for repo in self._fallback_repositories:
                # Writes don't affect fallback repos
                repo.lock_read()
            self._refresh_data()
        return RepositoryWriteLockResult(self.unlock, token)

    def lock_read(self):
        """Lock the repository for read operations.

        Returns: An object with an unlock method which will release the lock
            obtained.
        """
        locked = self.is_locked()
        self.control_files.lock_read()
        if not locked:
            self._warn_if_deprecated()
            self._note_lock('r')
            for repo in self._fallback_repositories:
                repo.lock_read()
            self._refresh_data()
        return LogicalLockResult(self.unlock)

    def get_physical_lock_status(self):
        return self.control_files.get_physical_lock_status()

    def leave_lock_in_place(self):
        """Tell this repository not to release the physical lock when this
        object is unlocked.

        If lock_write doesn't return a token, then this method is not supported.
        """
        self.control_files.leave_in_place()

    def dont_leave_lock_in_place(self):
        """Tell this repository to release the physical lock when this
        object is unlocked, even if it didn't originally acquire it.

        If lock_write doesn't return a token, then this method is not supported.
        """
        self.control_files.dont_leave_in_place()

    def gather_stats(self, revid=None, committers=None):
        """Gather statistics from a revision id.

        Args:
          revid: The revision id to gather statistics from, if None, then
            no revision specific statistics are gathered.
          committers: Optional parameter controlling whether to grab
            a count of committers from the revision specific statistics.
        Returns: A dictionary of statistics. Currently this contains:
            committers: The number of committers if requested.
            firstrev: A tuple with timestamp, timezone for the penultimate left
                most ancestor of revid, if revid is not the NULL_REVISION.
            latestrev: A tuple with timestamp, timezone for revid, if revid is
                not the NULL_REVISION.
            revisions: The total revision count in the repository.
            size: An estimate disk size of the repository in bytes.
        """
        with self.lock_read():
            result = {}
            if revid and committers:
                result['committers'] = 0
            if revid and revid != _mod_revision.NULL_REVISION:
                graph = self.get_graph()
                if committers:
                    all_committers = set()
                revisions = [r for (r, p) in graph.iter_ancestry([revid])
                             if r != _mod_revision.NULL_REVISION]
                last_revision = None
                if not committers:
                    # ignore the revisions in the middle - just grab first and last
                    revisions = revisions[0], revisions[-1]
                for revision in self.get_revisions(revisions):
                    if not last_revision:
                        last_revision = revision
                    if committers:
                        all_committers.add(revision.committer)
                first_revision = revision
                if committers:
                    result['committers'] = len(all_committers)
                result['firstrev'] = (first_revision.timestamp,
                                      first_revision.timezone)
                result['latestrev'] = (last_revision.timestamp,
                                       last_revision.timezone)
            return result

    def find_branches(self, using=False):
        """Find branches underneath this repository.

        This will include branches inside other branches.

        Args:
          using: If True, list only branches using this repository.
        """
        if using and not self.is_shared():
            yield from self.controldir.list_branches()
            return

        class Evaluator:

            def __init__(self):
                self.first_call = True

            def __call__(self, controldir):
                # On the first call, the parameter is always the controldir
                # containing the current repo.
                if not self.first_call:
                    try:
                        repository = controldir.open_repository()
                    except errors.NoRepositoryPresent:
                        pass
                    else:
                        return False, ([], repository)
                self.first_call = False
                value = (controldir.list_branches(), None)
                return True, value

        for branches, repository in controldir.ControlDir.find_controldirs(
                self.user_transport, evaluate=Evaluator()):
            if branches is not None:
                yield from branches
            if not using and repository is not None:
                yield from repository.find_branches()

    def search_missing_revision_ids(self, other,
                                    find_ghosts=True, revision_ids=None, if_present_ids=None,
                                    limit=None):
        """Return the revision ids that other has that this does not.

        These are returned in topological order.

        revision_ids: only return revision ids included by revision_id.
        """
        with self.lock_read():
            return InterRepository.get(other, self).search_missing_revision_ids(
                find_ghosts=find_ghosts, revision_ids=revision_ids,
                if_present_ids=if_present_ids, limit=limit)

    @staticmethod
    def open(base):
        """Open the repository rooted at base.

        For instance, if the repository is at URL/.bzr/repository,
        Repository.open(URL) -> a Repository instance.
        """
        control = controldir.ControlDir.open(base)
        return control.open_repository()

    def copy_content_into(self, destination, revision_id=None):
        """Make a complete copy of the content in self into destination.

        This is a destructive operation! Do not use it on existing
        repositories.
        """
        return InterRepository.get(self, destination).copy_content(revision_id)

    def commit_write_group(self):
        """Commit the contents accrued within the current write group.

        :seealso: start_write_group.

        Returns: it may return an opaque hint that can be passed to 'pack'.
        """
        if self._write_group is not self.get_transaction():
            # has an unlock or relock occured ?
            raise errors.BzrError('mismatched lock context %r and '
                                  'write group %r.' %
                                  (self.get_transaction(), self._write_group))
        result = self._commit_write_group()
        self._write_group = None
        return result

    def _commit_write_group(self):
        """Template method for per-repository write group cleanup.

        This is called before the write group is considered to be
        finished and should ensure that all data handed to the repository
        for writing during the write group is safely committed (to the
        extent possible considering file system caching etc).
        """

    def suspend_write_group(self):
        """Suspend a write group.

        Raises:
          UnsuspendableWriteGroup: If the write group can not be
            suspended.

        Returns:
          List of tokens
        """
        raise errors.UnsuspendableWriteGroup(self)

    def refresh_data(self):
        """Re-read any data needed to synchronise with disk.

        This method is intended to be called after another repository instance
        (such as one used by a smart server) has inserted data into the
        repository. On all repositories this will work outside of write groups.
        Some repository formats (pack and newer for breezy native formats)
        support refresh_data inside write groups. If called inside a write
        group on a repository that does not support refreshing in a write group
        IsInWriteGroupError will be raised.
        """
        self._refresh_data()

    def resume_write_group(self, tokens):
        if not self.is_write_locked():
            raise errors.NotWriteLocked(self)
        if self._write_group:
            raise errors.BzrError('already in a write group')
        self._resume_write_group(tokens)
        # so we can detect unlock/relock - the write group is now entered.
        self._write_group = self.get_transaction()

    def _resume_write_group(self, tokens):
        raise errors.UnsuspendableWriteGroup(self)

    def fetch(self, source, revision_id=None, find_ghosts=False, lossy=False):
        """Fetch the content required to construct revision_id from source.

        If revision_id is None, then all content is copied.

        fetch() may not be used when the repository is in a write group -
        either finish the current write group before using fetch, or use
        fetch before starting the write group.

        Args:
          find_ghosts: Find and copy revisions in the source that are
            ghosts in the target (and not reachable directly by walking out to
            the first-present revision in target from revision_id).
          revision_id: If specified, all the content needed for this
            revision ID will be copied to the target.  Fetch will determine for
            itself which content needs to be copied.
        Returns: A FetchResult object
        """
        if self.is_in_write_group():
            raise errors.InternalBzrError(
                "May not fetch while in a write group.")
        # fast path same-url fetch operations
        # TODO: lift out to somewhere common with RemoteRepository
        # <https://bugs.launchpad.net/bzr/+bug/401646>
        if (self.has_same_location(source)
                and self._has_same_fallbacks(source)):
            # check that last_revision is in 'from' and then return a
            # no-operation.
            if (revision_id is not None and
                    not _mod_revision.is_null(revision_id)):
                self.get_revision(revision_id)
            return 0, []
        inter = InterRepository.get(source, self)
        return inter.fetch(
            revision_id=revision_id, find_ghosts=find_ghosts, lossy=lossy)

    def get_commit_builder(self, branch, parents, config_stack, timestamp=None,
                           timezone=None, committer=None, revprops=None,
                           revision_id=None, lossy=False):
        """Obtain a CommitBuilder for this repository.

        Args:
          branch: Branch to commit to.
          parents: Revision ids of the parents of the new revision.
          config_stack: Configuration stack to use.
          timestamp: Optional timestamp recorded for commit.
          timezone: Optional timezone for timestamp.
          committer: Optional committer to set for commit.
          revprops: Optional dictionary of revision properties.
          revision_id: Optional revision id.
          lossy: Whether to discard data that can not be natively
            represented, when pushing to a foreign VCS
        """
        raise NotImplementedError(self.get_commit_builder)

    @only_raises(errors.LockNotHeld, errors.LockBroken)
    def unlock(self):
        if (self.control_files._lock_count == 1 and
                self.control_files._lock_mode == 'w'):
            if self._write_group is not None:
                self.abort_write_group()
                self.control_files.unlock()
                raise errors.BzrError(
                    'Must end write groups before releasing write locks.')
        self.control_files.unlock()
        if self.control_files._lock_count == 0:
            for repo in self._fallback_repositories:
                repo.unlock()

    def clone(self, controldir, revision_id=None):
        """Clone this repository into controldir using the current format.

        Currently no check is made that the format of this repository and
        the bzrdir format are compatible. FIXME RBC 20060201.

        Returns: The newly created destination repository.
        """
        with self.lock_read():
            # TODO: deprecate after 0.16; cloning this with all its settings is
            # probably not very useful -- mbp 20070423
            dest_repo = self._create_sprouting_repo(
                controldir, shared=self.is_shared())
            self.copy_content_into(dest_repo, revision_id)
            return dest_repo

    def start_write_group(self):
        """Start a write group in the repository.

        Write groups are used by repositories which do not have a 1:1 mapping
        between file ids and backend store to manage the insertion of data from
        both fetch and commit operations.

        A write lock is required around the
        start_write_group/commit_write_group for the support of lock-requiring
        repository formats.

        One can only insert data into a repository inside a write group.

        Returns: None.
        """
        if not self.is_write_locked():
            raise errors.NotWriteLocked(self)
        if self._write_group:
            raise errors.BzrError('already in a write group')
        self._start_write_group()
        # so we can detect unlock/relock - the write group is now entered.
        self._write_group = self.get_transaction()

    def _start_write_group(self):
        """Template method for per-repository write group startup.

        This is called before the write group is considered to be
        entered.
        """

    def sprout(self, to_bzrdir, revision_id=None):
        """Create a descendent repository for new development.

        Unlike clone, this does not copy the settings of the repository.
        """
        with self.lock_read():
            dest_repo = self._create_sprouting_repo(to_bzrdir, shared=False)
            dest_repo.fetch(self, revision_id=revision_id)
            return dest_repo

    def _create_sprouting_repo(self, a_controldir, shared):
        if not isinstance(
                a_controldir._format, self.controldir._format.__class__):
            # use target default format.
            dest_repo = a_controldir.create_repository()
        else:
            # Most control formats need the repository to be specifically
            # created, but on some old all-in-one formats it's not needed
            try:
                dest_repo = self._format.initialize(
                    a_controldir, shared=shared)
            except errors.UninitializableFormat:
                dest_repo = a_controldir.open_repository()
        return dest_repo

    def has_revision(self, revision_id):
        """True if this repository has a copy of the revision."""
        with self.lock_read():
            return revision_id in self.has_revisions((revision_id,))

    def has_revisions(self, revision_ids):
        """Probe to find out the presence of multiple revisions.

        Args:
          revision_ids: An iterable of revision_ids.
        Returns: A set of the revision_ids that were present.
        """
        raise NotImplementedError(self.has_revisions)

    def get_revision(self, revision_id):
        """Return the Revision object for a named revision."""
        with self.lock_read():
            return self.get_revisions([revision_id])[0]

    def get_revision_reconcile(self, revision_id):
        """'reconcile' helper routine that allows access to a revision always.

        This variant of get_revision does not cross check the weave graph
        against the revision one as get_revision does: but it should only
        be used by reconcile, or reconcile-alike commands that are correcting
        or testing the revision graph.
        """
        raise NotImplementedError(self.get_revision_reconcile)

    def get_revisions(self, revision_ids):
        """Get many revisions at once.

        Repositories that need to check data on every revision read should
        subclass this method.
        """
        revs = {}
        for revid, rev in self.iter_revisions(revision_ids):
            if rev is None:
                raise errors.NoSuchRevision(self, revid)
            revs[revid] = rev
        return [revs[revid] for revid in revision_ids]

    def iter_revisions(self, revision_ids):
        """Iterate over revision objects.

        Args:
          revision_ids: An iterable of revisions to examine. None may be
            passed to request all revisions known to the repository. Note that
            not all repositories can find unreferenced revisions; for those
            repositories only referenced ones will be returned.
        Returns: An iterator of (revid, revision) tuples. Absent revisions (
            those asked for but not available) are returned as (revid, None).
            N.B.: Revisions are not necessarily yielded in order.
        """
        raise NotImplementedError(self.iter_revisions)

    def get_revision_delta(self, revision_id):
        """Return the delta for one revision.

        The delta is relative to the left-hand predecessor of the
        revision.
        """
        with self.lock_read():
            r = self.get_revision(revision_id)
            return list(self.get_revision_deltas([r]))[0]

    def get_revision_deltas(self, revisions, specific_files=None):
        """Produce a generator of revision deltas.

        Note that the input is a sequence of REVISIONS, not revision ids.
        Trees will be held in memory until the generator exits.
        Each delta is relative to the revision's lefthand predecessor.

        specific_files should exist in the first revision.

        Args:
          specific_files: if not None, the result is filtered
          so that only those files, their parents and their
          children are included.
        """
        from .tree import InterTree

        # Get the revision-ids of interest
        required_trees = set()
        for revision in revisions:
            required_trees.add(revision.revision_id)
            required_trees.update(revision.parent_ids[:1])

        trees = {
            t.get_revision_id(): t
            for t in self.revision_trees(required_trees)}

        # Calculate the deltas
        for revision in revisions:
            if not revision.parent_ids:
                old_tree = self.revision_tree(_mod_revision.NULL_REVISION)
            else:
                old_tree = trees[revision.parent_ids[0]]
            intertree = InterTree.get(old_tree, trees[revision.revision_id])
            yield intertree.compare(specific_files=specific_files)
            if specific_files is not None:
                specific_files = [
                    p for p in intertree.find_source_paths(
                        specific_files).values()
                    if p is not None]

    def store_revision_signature(self, gpg_strategy, plaintext, revision_id):
        raise NotImplementedError(self.store_revision_signature)

    def add_signature_text(self, revision_id, signature):
        """Store a signature text for a revision.

        Args:
          revision_id: Revision id of the revision
          signature: Signature text.
        """
        raise NotImplementedError(self.add_signature_text)

    def iter_files_bytes(self, desired_files):
        """Iterate through file versions.

        Files will not necessarily be returned in the order they occur in
        desired_files.  No specific order is guaranteed.

        Yields pairs of identifier, bytes_iterator.  identifier is an opaque
        value supplied by the caller as part of desired_files.  It should
        uniquely identify the file version in the caller's context.  (Examples:
        an index number or a TreeTransform trans_id.)

        Args:
          desired_files: a list of (file_id, revision_id, identifier)
            triples
        """
        raise NotImplementedError(self.iter_files_bytes)

    def get_rev_id_for_revno(self, revno, known_pair):
        """Return the revision id of a revno, given a later (revno, revid)
        pair in the same history.

        Returns: if found (True, revid).  If the available history ran out
            before reaching the revno, then this returns
            (False, (closest_revno, closest_revid)).
        """
        known_revno, known_revid = known_pair
        partial_history = [known_revid]
        distance_from_known = known_revno - revno
        if distance_from_known < 0:
            raise errors.RevnoOutOfBounds(revno, (0, known_revno))
        try:
            _iter_for_revno(
                self, partial_history, stop_index=distance_from_known)
        except errors.RevisionNotPresent as err:
            if err.revision_id == known_revid:
                # The start revision (known_revid) wasn't found.
                raise errors.NoSuchRevision(self, known_revid)
            # This is a stacked repository with no fallbacks, or a there's a
            # left-hand ghost.  Either way, even though the revision named in
            # the error isn't in this repo, we know it's the next step in this
            # left-hand history.
            partial_history.append(err.revision_id)
        if len(partial_history) <= distance_from_known:
            # Didn't find enough history to get a revid for the revno.
            earliest_revno = known_revno - len(partial_history) + 1
            return (False, (earliest_revno, partial_history[-1]))
        if len(partial_history) - 1 > distance_from_known:
            raise AssertionError('_iter_for_revno returned too much history')
        return (True, partial_history[-1])

    def is_shared(self):
        """Return True if this repository is flagged as a shared repository."""
        raise NotImplementedError(self.is_shared)

    def reconcile(self, other=None, thorough=False):
        """Reconcile this repository."""
        raise NotImplementedError(self.reconcile)

    def _refresh_data(self):
        """Helper called from lock_* to ensure coherency with disk.

        The default implementation does nothing; it is however possible
        for repositories to maintain loaded indices across multiple locks
        by checking inside their implementation of this method to see
        whether their indices are still valid. This depends of course on
        the disk format being validatable in this manner. This method is
        also called by the refresh_data() public interface to cause a refresh
        to occur while in a write lock so that data inserted by a smart server
        push operation is visible on the client's instance of the physical
        repository.
        """

    def revision_tree(self, revision_id) -> "RevisionTree":
        """Return Tree for a revision on this branch.

        `revision_id` may be NULL_REVISION for the empty tree revision.
        """
        raise NotImplementedError(self.revision_tree)

    def revision_trees(self, revision_ids):
        """Return Trees for revisions in this repository.

        Args:
          revision_ids: a sequence of revision-ids;
          a revision-id may not be None or b'null:'
        """
        raise NotImplementedError(self.revision_trees)

    def pack(self, hint=None, clean_obsolete_packs=False):
        """Compress the data within the repository.

        This operation only makes sense for some repository types. For other
        types it should be a no-op that just returns.

        This stub method does not require a lock, but subclasses should use
        self.write_lock as this is a long running call it's reasonable to
        implicitly lock for the user.

        Args:
          hint: If not supplied, the whole repository is packed.
            If supplied, the repository may use the hint parameter as a
            hint for the parts of the repository to pack. A hint can be
            obtained from the result of commit_write_group(). Out of
            date hints are simply ignored, because concurrent operations
            can obsolete them rapidly.

          clean_obsolete_packs: Clean obsolete packs immediately after
            the pack operation.
        """

    def get_transaction(self):
        return self.control_files.get_transaction()

    def get_parent_map(self, revision_ids):
        """See graph.StackedParentsProvider.get_parent_map"""
        raise NotImplementedError(self.get_parent_map)

    def _get_parent_map_no_fallbacks(self, revision_ids):
        """Same as Repository.get_parent_map except doesn't query fallbacks."""
        # revisions index works in keys; this just works in revisions
        # therefore wrap and unwrap
        query_keys = []
        result = {}
        for revision_id in revision_ids:
            if revision_id == _mod_revision.NULL_REVISION:
                result[revision_id] = ()
            elif revision_id is None:
                raise ValueError('get_parent_map(None) is not valid')
            else:
                query_keys.append((revision_id,))
        vf = self.revisions.without_fallbacks()
        for (revision_id,), parent_keys in (
                vf.get_parent_map(query_keys).items()):
            if parent_keys:
                result[revision_id] = tuple([parent_revid
                                             for (parent_revid,) in parent_keys])
            else:
                result[revision_id] = (_mod_revision.NULL_REVISION,)
        return result

    def _make_parents_provider(self):
        if not self._format.supports_external_lookups:
            return self
        return graph.StackedParentsProvider(_LazyListJoin(
            [self._make_parents_provider_unstacked()],
            self._fallback_repositories))

    def _make_parents_provider_unstacked(self):
        return graph.CallableToParentsProviderAdapter(
            self._get_parent_map_no_fallbacks)

    def get_known_graph_ancestry(self, revision_ids):
        """Return the known graph for a set of revision ids and their ancestors.
        """
        raise NotImplementedError(self.get_known_graph_ancestry)

    def get_file_graph(self):
        """Return the graph walker for files."""
        raise NotImplementedError(self.get_file_graph)

    def get_graph(self, other_repository=None):
        """Return the graph walker for this repository format"""
        parents_provider = self._make_parents_provider()
        if (other_repository is not None and
                not self.has_same_location(other_repository)):
            parents_provider = graph.StackedParentsProvider(
                [parents_provider, other_repository._make_parents_provider()])
        return graph.Graph(parents_provider)

    def set_make_working_trees(self, new_value):
        """Set the policy flag for making working trees when creating branches.

        This only applies to branches that use this repository.

        The default is 'True'.

        Args:
          new_value: True to restore the default, False to disable making
                          working trees.
        """
        raise NotImplementedError(self.set_make_working_trees)

    def make_working_trees(self):
        """Returns the policy for making working trees on new branches."""
        raise NotImplementedError(self.make_working_trees)

    def sign_revision(self, revision_id, gpg_strategy):
        raise NotImplementedError(self.sign_revision)

    def verify_revision_signature(self, revision_id, gpg_strategy):
        """Verify the signature on a revision.

        Args:
          revision_id: the revision to verify
          gpg_strategy: the GPGStrategy object to used

        Returns: gpg.SIGNATURE_VALID or a failed SIGNATURE_ value
        """
        raise NotImplementedError(self.verify_revision_signature)

    def verify_revision_signatures(self, revision_ids, gpg_strategy):
        """Verify revision signatures for a number of revisions.

        Args:
          revision_id: the revision to verify
          gpg_strategy: the GPGStrategy object to used

        Returns:
          Iterator over tuples with revision id, result and keys
        """
        with self.lock_read():
            for revid in revision_ids:
                (result, key) = self.verify_revision_signature(revid, gpg_strategy)
                yield revid, result, key

    def has_signature_for_revision_id(self, revision_id):
        """Query for a revision signature for revision_id in the repository."""
        raise NotImplementedError(self.has_signature_for_revision_id)

    def get_signature_text(self, revision_id):
        """Return the text for a signature."""
        raise NotImplementedError(self.get_signature_text)

    def check(self, revision_ids=None, callback_refs=None, check_repo=True):
        """Check consistency of all history of given revision_ids.

        Different repository implementations should override _check().

        Args:
          revision_ids: A non-empty list of revision_ids whose ancestry
             will be checked.  Typically the last revision_id of a branch.
          callback_refs: A dict of check-refs to resolve and callback
            the check/_check method on the items listed as wanting the ref.
            see breezy.check.
          check_repo: If False do not check the repository contents, just
            calculate the data callback_refs requires and call them back.
        """
        return self._check(revision_ids=revision_ids, callback_refs=callback_refs,
                           check_repo=check_repo)

    def _check(self, revision_ids=None, callback_refs=None, check_repo=True):
        raise NotImplementedError(self.check)

    def _warn_if_deprecated(self, branch=None):
        if not self._format.is_deprecated():
            return
        global _deprecation_warning_done
        if _deprecation_warning_done:
            return
        try:
            if branch is None:
                conf = config.GlobalStack()
            else:
                conf = branch.get_config_stack()
            if 'format_deprecation' in conf.get('suppress_warnings'):
                return
            warning("Format %s for %s is deprecated -"
                    " please use 'brz upgrade' to get better performance"
                    % (self._format, self.controldir.transport.base))
        finally:
            _deprecation_warning_done = True

    def supports_rich_root(self):
        return self._format.rich_root_data

    def _check_ascii_revisionid(self, revision_id, method):
        """Private helper for ascii-only repositories."""
        # weave repositories refuse to store revisionids that are non-ascii.
        if revision_id is not None:
            # weaves require ascii revision ids.
            if isinstance(revision_id, str):
                try:
                    revision_id.encode('ascii')
                except UnicodeEncodeError:
                    raise errors.NonAsciiRevisionId(method, self)
            else:
                try:
                    revision_id.decode('ascii')
                except UnicodeDecodeError:
                    raise errors.NonAsciiRevisionId(method, self)


class RepositoryFormatRegistry(controldir.ControlComponentFormatRegistry):
    """Repository format registry."""

    def get_default(self):
        """Return the current default format."""
        return controldir.format_registry.make_controldir('default').repository_format


network_format_registry = registry.FormatRegistry["RepositoryFormat"]()
"""Registry of formats indexed by their network name.

The network name for a repository format is an identifier that can be used when
referring to formats with smart server operations. See
RepositoryFormat.network_name() for more detail.
"""


format_registry = RepositoryFormatRegistry(network_format_registry)
"""Registry of formats, indexed by their BzrDirMetaFormat format string.

This can contain either format instances themselves, or classes/factories that
can be called to obtain one.
"""


#####################################################################
# Repository Formats

class RepositoryFormat(controldir.ControlComponentFormat):
    """A repository format.

    Formats provide four things:
     * An initialization routine to construct repository data on disk.
     * a optional format string which is used when the BzrDir supports
       versioned children.
     * an open routine which returns a Repository instance.
     * A network name for referring to the format in smart server RPC
       methods.

    There is one and only one Format subclass for each on-disk format. But
    there can be one Repository subclass that is used for several different
    formats. The _format attribute on a Repository instance can be used to
    determine the disk format.

    Formats are placed in a registry by their format string for reference
    during opening. These should be subclasses of RepositoryFormat for
    consistency.

    Once a format is deprecated, just deprecate the initialize and open
    methods on the format class. Do not deprecate the object, as the
    object may be created even when a repository instance hasn't been
    created.

    Common instance attributes:
    _matchingcontroldir - the controldir format that the repository format was
    originally written to work with. This can be used if manually
    constructing a bzrdir and repository, or more commonly for test suite
    parameterization.
    """

    # Set to True or False in derived classes. True indicates that the format
    # supports ghosts gracefully.
    supports_ghosts: bool
    # Can this repository be given external locations to lookup additional
    # data. Set to True or False in derived classes.
    supports_external_lookups: bool
    # Does this format support CHK bytestring lookups. Set to True or False in
    # derived classes.
    supports_chks: bool
    # Should fetch trigger a reconcile after the fetch? Only needed for
    # some repository formats that can suffer internal inconsistencies.
    _fetch_reconcile: bool = False
    # Does this format have < O(tree_size) delta generation. Used to hint what
    # code path for commit, amongst other things.
    fast_deltas: bool
    # Does doing a pack operation compress data? Useful for the pack UI command
    # (so if there is one pack, the operation can still proceed because it may
    # help), and for fetching when data won't have come from the same
    # compressor.
    pack_compresses: bool = False
    # Does the repository storage understand references to trees?
    supports_tree_reference: bool
    # Is the format experimental ?
    experimental: bool = False
    # Does this repository format escape funky characters, or does it create
    # files with similar names as the versioned files in its contents on disk
    # ?
    supports_funky_characters: bool
    # Does this repository format support leaving locks?
    supports_leaving_lock: bool
    # Does this format support the full VersionedFiles interface?
    supports_full_versioned_files: bool
    # Does this format support signing revision signatures?
    supports_revision_signatures: bool = True
    # Can the revision graph have incorrect parents?
    revision_graph_can_have_wrong_parents: bool
    # Does this format support setting revision ids?
    supports_setting_revision_ids: bool = True
    # Does this format support rich root data?
    rich_root_data: bool
    # Does this format support explicitly versioned directories?
    supports_versioned_directories: bool
    # Can other repositories be nested into one of this format?
    supports_nesting_repositories: bool
    # Is it possible for revisions to be present without being referenced
    # somewhere ?
    supports_unreferenced_revisions: bool
    # Does this format store the current Branch.nick in a revision when
    # creating commits?
    supports_storing_branch_nick: bool = True
    # Does the format support overriding the transport to use
    supports_overriding_transport: bool = True
    # Does the format support setting custom revision properties?
    supports_custom_revision_properties: bool = True
    # Does the format record per-file revision metadata?
    records_per_file_revision: bool = True
    supports_multiple_authors: bool = True

    def __repr__(self):
        return "%s()" % self.__class__.__name__

    def __eq__(self, other):
        # format objects are generally stateless
        return isinstance(other, self.__class__)

    def __ne__(self, other):
        return not self == other

    def get_format_description(self):
        """Return the short description for this format."""
        raise NotImplementedError(self.get_format_description)

    def initialize(self, controldir, shared=False):
        """Initialize a repository of this format in controldir.

        Args:
          controldir: The controldir to put the new repository in it.
          shared: The repository should be initialized as a sharable one.

        Returns:
          The new repository object.

        This may raise UninitializableFormat if shared repository are not
        compatible the controldir.
        """
        raise NotImplementedError(self.initialize)

    def is_supported(self):
        """Is this format supported?

        Supported formats must be initializable and openable.
        Unsupported formats may not support initialization or committing or
        some other features depending on the reason for not being supported.
        """
        return True

    def is_deprecated(self):
        """Is this format deprecated?

        Deprecated formats may trigger a user-visible warning recommending
        the user to upgrade. They are still fully supported.
        """
        return False

    def network_name(self):
        """A simple byte string uniquely identifying this format for RPC calls.

        MetaDir repository formats use their disk format string to identify the
        repository over the wire. All in one formats such as bzr < 0.8, and
        foreign formats like svn/git and hg should use some marker which is
        unique and immutable.
        """
        raise NotImplementedError(self.network_name)

    def check_conversion_target(self, target_format):
        if self.rich_root_data and not target_format.rich_root_data:
            raise errors.BadConversionTarget(
                'Does not support rich root data.', target_format,
                from_format=self)
        if (self.supports_tree_reference
                and not getattr(target_format, 'supports_tree_reference', False)):
            raise errors.BadConversionTarget(
                'Does not support nested trees', target_format,
                from_format=self)

    def open(self, controldir, _found=False):
        """Return an instance of this format for a controldir.

        _found is a private parameter, do not use it.
        """
        raise NotImplementedError(self.open)

    def _run_post_repo_init_hooks(self, repository, controldir, shared):
        from .controldir import ControlDir, RepoInitHookParams
        hooks = ControlDir.hooks['post_repo_init']
        if not hooks:
            return
        params = RepoInitHookParams(repository, self, controldir, shared)
        for hook in hooks:
            hook(params)


class AbstractSearchResult:
    """The result of a search, describing a set of keys.

    Search results are typically used as the 'fetch_spec' parameter when
    fetching revisions.

    :seealso: AbstractSearch
    """

    def get_recipe(self):
        """Return a recipe that can be used to replay this search.

        The recipe allows reconstruction of the same results at a later date.

        :return: A tuple of `(search_kind_str, *details)`.  The details vary by
            kind of search result.
        """
        raise NotImplementedError(self.get_recipe)

    def get_network_struct(self):
        """Return a tuple that can be transmitted via the HPSS protocol."""
        raise NotImplementedError(self.get_network_struct)

    def get_keys(self):
        """Return the keys found in this search.

        :return: A set of keys.
        """
        raise NotImplementedError(self.get_keys)

    def is_empty(self):
        """Return false if the search lists 1 or more revisions."""
        raise NotImplementedError(self.is_empty)

    def refine(self, seen, referenced):
        """Create a new search by refining this search.

        :param seen: Revisions that have been satisfied.
        :param referenced: Revision references observed while satisfying some
            of this search.
        :return: A search result.
        """
        raise NotImplementedError(self.refine)


class InterRepository(InterObject[Repository]):
    """This class represents operations taking place between two repositories.

    Its instances have methods like copy_content and fetch, and contain
    references to the source and target repositories these operations can be
    carried out on.

    Often we will provide convenience methods on 'repository' which carry out
    operations with another repository - they will always forward to
    InterRepository.get(other).method_name(parameters).
    """

    _optimisers = []
    """The available optimised InterRepository types."""

    def copy_content(self, revision_id: Optional[_mod_revision.RevisionID] = None) -> None:
        """Make a complete copy of the content in self into destination.

        This is a destructive operation! Do not use it on existing
        repositories.

        Args:
          revision_id: Only copy the content needed to construct
                            revision_id and its parents.
        """
        with self.lock_write():
            try:
                self.target.set_make_working_trees(
                    self.source.make_working_trees())
            except (NotImplementedError, errors.RepositoryUpgradeRequired):
                pass
            self.target.fetch(self.source, revision_id=revision_id)

    def fetch(self, revision_id: Optional[_mod_revision.RevisionID] = None, find_ghosts: bool = False, lossy: bool = False) -> FetchResult:
        """Fetch the content required to construct revision_id.

        The content is copied from self.source to self.target.

        Args:
          revision_id: if None all content is copied, if NULL_REVISION no
                            content is copied.
        Returns: FetchResult
        """
        raise NotImplementedError(self.fetch)

    def search_missing_revision_ids(
            self, find_ghosts: bool = True, revision_ids: Optional[Iterable[_mod_revision.RevisionID]] = None,
            if_present_ids: Optional[Iterable[_mod_revision.RevisionID]] = None,
            limit: Optional[int] = None) -> AbstractSearchResult:
        """Return the revision ids that source has that target does not.

        Args:
          revision_ids: return revision ids included by these
            revision_ids.  NoSuchRevision will be raised if any of these
            revisions are not present.
          if_present_ids: like revision_ids, but will not cause
            NoSuchRevision if any of these are absent, instead they will simply
            not be in the result.  This is useful for e.g. finding revisions
            to fetch for tags, which may reference absent revisions.
          find_ghosts: If True find missing revisions in deep history
            rather than just finding the surface difference.
          limit: Maximum number of revisions to return, topologically
            ordered
        Returns: A SearchResult.
        """
        raise NotImplementedError(self.search_missing_revision_ids)

    @staticmethod
    def _same_model(source, target):
        """True if source and target have the same data representation.

        Note: this is always called on the base class; overriding it in a
        subclass will have no effect.
        """
        try:
            InterRepository._assert_same_model(source, target)
            return True
        except errors.IncompatibleRepositories as e:
            return False

    @staticmethod
    def _assert_same_model(source, target):
        """Raise an exception if two repositories do not use the same model.
        """
        if source.supports_rich_root() != target.supports_rich_root():
            raise errors.IncompatibleRepositories(source, target,
                                                  "different rich-root support")
        if not hasattr(source, '_serializer') or not hasattr(target, '_serializer'):
            if source != target:
                raise errors.IncompatibleRepositories(source, target, "different formats")
            return

        if source._serializer != target._serializer:
            raise errors.IncompatibleRepositories(source, target,
                                                  "different serializers")


class CopyConverter:
    """A repository conversion tool which just performs a copy of the content.

    This is slow but quite reliable.
    """

    def __init__(self, target_format):
        """Create a CopyConverter.

        Args:
          target_format: The format the resulting repository should be.
        """
        self.target_format = target_format

    def convert(self, repo, pb):
        """Perform the conversion of to_convert, giving feedback via pb.

        Args:
          to_convert: The disk object to convert.
          pb: a progress bar to use for progress information.
        """
        with ui.ui_factory.nested_progress_bar() as pb:
            self.count = 0
            self.total = 4
            # this is only useful with metadir layouts - separated repo content.
            # trigger an assertion if not such
            repo._format.get_format_string()
            self.repo_dir = repo.controldir
            pb.update(gettext('Moving repository to repository.backup'))
            self.repo_dir.transport.move('repository', 'repository.backup')
            backup_transport = self.repo_dir.transport.clone(
                'repository.backup')
            repo._format.check_conversion_target(self.target_format)
            self.source_repo = repo._format.open(self.repo_dir,
                                                 _found=True,
                                                 _override_transport=backup_transport)
            pb.update(gettext('Creating new repository'))
            converted = self.target_format.initialize(self.repo_dir,
                                                      self.source_repo.is_shared())
            with converted.lock_write():
                pb.update(gettext('Copying content'))
                self.source_repo.copy_content_into(converted)
            pb.update(gettext('Deleting old repository content'))
            self.repo_dir.transport.delete_tree('repository.backup')
            ui.ui_factory.note(gettext('repository converted'))


def _strip_NULL_ghosts(revision_graph):
    """Also don't use this. more compatibility code for unmigrated clients."""
    # Filter ghosts, and null:
    if _mod_revision.NULL_REVISION in revision_graph:
        del revision_graph[_mod_revision.NULL_REVISION]
    for key, parents in revision_graph.items():
        revision_graph[key] = tuple(parent for parent in parents if parent
                                    in revision_graph)
    return revision_graph


def _iter_for_revno(repo, partial_history_cache, stop_index=None,
                    stop_revision=None):
    """Extend the partial history to include a given index

    If a stop_index is supplied, stop when that index has been reached.
    If a stop_revision is supplied, stop when that revision is
    encountered.  Otherwise, stop when the beginning of history is
    reached.

    Args:
      stop_index: The index which should be present.  When it is
        present, history extension will stop.
      stop_revision: The revision id which should be present.  When
        it is encountered, history extension will stop.
    """
    start_revision = partial_history_cache[-1]
    graph = repo.get_graph()
    iterator = graph.iter_lefthand_ancestry(start_revision,
                                            (_mod_revision.NULL_REVISION,))
    try:
        # skip the last revision in the list
        next(iterator)
        while True:
            if (stop_index is not None and
                    len(partial_history_cache) > stop_index):
                break
            if partial_history_cache[-1] == stop_revision:
                break
            revision_id = next(iterator)
            partial_history_cache.append(revision_id)
    except StopIteration:
        # No more history
        return


class _LazyListJoin:
    """An iterable yielding the contents of many lists as one list.

    Each iterator made from this will reflect the current contents of the lists
    at the time the iterator is made.

    This is used by Repository's _make_parents_provider implementation so that
    it is safe to do::

      pp = repo._make_parents_provider()      # uses a list of fallback repos
      pp.add_fallback_repository(other_repo)  # appends to that list
      result = pp.get_parent_map(...)
      # The result will include revs from other_repo
    """

    def __init__(self, *list_parts):
        self.list_parts = list_parts

    def __iter__(self):
        full_list = []
        for list_part in self.list_parts:
            full_list.extend(list_part)
        return iter(full_list)

    def __repr__(self):
        return "{}.{}({})".format(self.__module__, self.__class__.__name__,
                              self.list_parts)



# Copyright (C) 2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for url handling."""

from ...tests import TestCase
from ..urls import git_url_to_bzr_url


class TestConvertURL(TestCase):

    def test_simple(self):
        self.assertEqual(
            git_url_to_bzr_url('foo:bar/path'),
            'git+ssh://foo/bar/path')
        self.assertEqual(
            git_url_to_bzr_url(
                'user@foo:bar/path'),
            ('git+ssh://user@foo/bar/path'))

    def test_regular(self):
        self.assertEqual(
            git_url_to_bzr_url(
                'git+ssh://user@foo/bar/path'),
            ('git+ssh://user@foo/bar/path'))

    def test_just_ssh(self):
        self.assertEqual(
            git_url_to_bzr_url(
                'ssh://user@foo/bar/path'),
            ('git+ssh://user@foo/bar/path'))

    def test_path(self):
        self.assertEqual(git_url_to_bzr_url('/bar/path'), ('/bar/path'))

    def test_with_ref(self):
        self.assertEqual(
            git_url_to_bzr_url('foo:bar/path', ref=b'HEAD'),
            'git+ssh://foo/bar/path')
        self.assertEqual(
            git_url_to_bzr_url('foo:bar/path', ref=b'refs/heads/blah'),
            'git+ssh://foo/bar/path,branch=blah')
        self.assertEqual(
            git_url_to_bzr_url('foo:bar/path', ref=b'refs/tags/blah'),
            'git+ssh://foo/bar/path,ref=refs%2Ftags%2Fblah')

    def test_with_branch(self):
        self.assertEqual(
            git_url_to_bzr_url('foo:bar/path', branch=''),
            'git+ssh://foo/bar/path')
        self.assertEqual(
            git_url_to_bzr_url('foo:bar/path', branch='foo/blah'),
            'git+ssh://foo/bar/path,branch=foo%2Fblah')
        self.assertEqual(
            git_url_to_bzr_url('foo:bar/path', branch='blah'),
            'git+ssh://foo/bar/path,branch=blah')



# Copyright (C) 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for the InterTree.file_content_matches() function."""

from breezy.tests.per_intertree import TestCaseWithTwoTrees


class TestFileContentMatches(TestCaseWithTwoTrees):

    def test_same_contents_and_verifier(self):
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        self.build_tree_contents([
            ('1/file', b'apples'),
            ('2/file', b'apples'),
            ])
        tree1.add('file', ids=b'file-id-1')
        tree2.add('file', ids=b'file-id-2')
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        inter = self.intertree_class(tree1, tree2)
        self.assertTrue(inter.file_content_matches('file', 'file'))

    def test_different_contents_and_same_verifier(self):
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        self.build_tree_contents([
            ('1/file', b'apples'),
            ('2/file', b'oranges'),
            ])
        tree1.add('file', ids=b'file-id-1')
        tree2.add('file', ids=b'file-id-2')
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        inter = self.intertree_class(tree1, tree2)
        self.assertFalse(inter.file_content_matches(
            'file', 'file'))



# Copyright (C) 2009-2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Support for committing in native Git working trees."""

import stat

from dulwich.index import commit_tree, read_submodule_head
from dulwich.objects import Blob, Commit

from .. import bugtracker
from .. import config as _mod_config
from .. import gpg, osutils
from .. import revision as _mod_revision
from .. import trace
from ..errors import BzrError, RootMissing, UnsupportedOperation
from ..repository import CommitBuilder
from .mapping import encode_git_path, fix_person_identifier, object_mode
from .tree import entry_factory


class GitCommitBuilder(CommitBuilder):
    """Commit builder for Git repositories."""

    supports_record_entry_contents = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.random_revid = True
        self._validate_revprops(self._revprops)
        self.store = self.repository._git.object_store
        self._blobs = {}
        self._inv_delta = []
        self._deleted_paths = set()
        self._any_changes = False
        self._mapping = self.repository.get_mapping()

    def any_changes(self):
        return self._any_changes

    def record_iter_changes(self, workingtree, basis_revid, iter_changes):
        seen_root = False
        for change in iter_changes:
            if change.kind == (None, None):
                # Ephemeral
                continue
            if change.versioned[0] and not change.copied:
                file_id = self._mapping.generate_file_id(change.path[0])
            elif change.versioned[1]:
                file_id = self._mapping.generate_file_id(change.path[1])
            else:
                file_id = None
            if change.path[1]:
                parent_id_new = self._mapping.generate_file_id(osutils.dirname(change.path[1]))
            else:
                parent_id_new = None
            if change.kind[1] in ("directory",):
                self._inv_delta.append(
                    (change.path[0], change.path[1], file_id,
                     entry_factory[change.kind[1]](
                         file_id, change.name[1], parent_id_new)))
                if change.kind[0] in ("file", "symlink"):
                    self._blobs[encode_git_path(change.path[0])] = None
                    self._any_changes = True
                if change.path[1] == "":
                    seen_root = True
                continue
            self._any_changes = True
            if change.path[1] is None:
                self._inv_delta.append((change.path[0], change.path[1], file_id, None))
                self._deleted_paths.add(encode_git_path(change.path[0]))
                continue
            try:
                entry_kls = entry_factory[change.kind[1]]
            except KeyError:
                raise KeyError("unknown kind %s" % change.kind[1])
            entry = entry_kls(file_id, change.name[1], parent_id_new)
            if change.kind[1] == "file":
                entry.executable = change.executable[1]
                blob = Blob()
                f, st = workingtree.get_file_with_stat(change.path[1])
                try:
                    blob.data = f.read()
                finally:
                    f.close()
                sha = blob.id
                if st is not None:
                    entry.text_size = st.st_size
                else:
                    entry.text_size = len(blob.data)
                entry.git_sha1 = sha
                self.store.add_object(blob)
            elif change.kind[1] == "symlink":
                symlink_target = workingtree.get_symlink_target(change.path[1])
                blob = Blob()
                blob.data = encode_git_path(symlink_target)
                self.store.add_object(blob)
                sha = blob.id
                entry.symlink_target = symlink_target
                st = None
            elif change.kind[1] == "tree-reference":
                sha = read_submodule_head(workingtree.abspath(change.path[1]))
                reference_revision = workingtree.get_reference_revision(change.path[1])
                entry.reference_revision = reference_revision
                st = None
            else:
                raise AssertionError("Unknown kind %r" % change.kind[1])
            mode = object_mode(change.kind[1], change.executable[1])
            self._inv_delta.append((change.path[0], change.path[1], file_id, entry))
            if change.path[0] is not None:
                self._deleted_paths.add(encode_git_path(change.path[0]))
            self._blobs[encode_git_path(change.path[1])] = (mode, sha)
            if st is not None:
                yield change.path[1], (entry.git_sha1, st)
        if not seen_root and len(self.parents) == 0:
            raise RootMissing()
        if getattr(workingtree, "basis_tree", False):
            basis_tree = workingtree.basis_tree()
        else:
            if len(self.parents) == 0:
                basis_revid = _mod_revision.NULL_REVISION
            else:
                basis_revid = self.parents[0]
            basis_tree = self.repository.revision_tree(basis_revid)
        # Fill in entries that were not changed
        for entry in basis_tree._iter_tree_contents(include_trees=False):
            if entry.path in self._blobs:
                continue
            if entry.path in self._deleted_paths:
                continue
            self._blobs[entry.path] = (entry.mode, entry.sha)
        self.new_inventory = None

    def update_basis(self, tree):
        # Nothing to do here
        pass

    def finish_inventory(self):
        # eliminate blobs that were removed
        self._blobs = {k: v for (k, v) in self._blobs.items()}

    def _iterblobs(self):
        return ((path, sha, mode) for (path, (mode, sha))
                in self._blobs.items())

    def commit(self, message):
        self._validate_unicode_text(message, 'commit message')
        c = Commit()
        c.parents = [self.repository.lookup_bzr_revision_id(
            revid)[0] for revid in self.parents]
        c.tree = commit_tree(self.store, self._iterblobs())
        encoding = self._revprops.pop('git-explicit-encoding', 'utf-8')
        c.encoding = encoding.encode('ascii')
        c.committer = fix_person_identifier(self._committer.encode(encoding))
        pseudoheaders = []
        try:
            author = self._revprops.pop('author')
        except KeyError:
            try:
                authors = self._revprops.pop('authors').splitlines()
            except KeyError:
                author = self._committer
            else:
                if len(authors) == 0:
                    author = self._committer
                else:
                    author = authors[0]
                    for coauthor in authors[1:]:
                        pseudoheaders.append(
                            b'Co-authored-by: %s'
                            % fix_person_identifier(coauthor.encode(encoding)))
        c.author = fix_person_identifier(author.encode(encoding))
        bugstext = self._revprops.pop('bugs', None)
        if bugstext is not None:
            for url, status in bugtracker.decode_bug_urls(bugstext):
                if status == bugtracker.FIXED:
                    pseudoheaders.append(("Fixes: %s" % url).encode(encoding))
                elif status == bugtracker.RELATED:
                    pseudoheaders.append(("Bug: %s" % url).encode(encoding))
                else:
                    raise bugtracker.InvalidBugStatus(status)
        if self._revprops:
            raise NotImplementedError(self._revprops)
        c.commit_time = int(self._timestamp)
        c.author_time = int(self._timestamp)
        c.commit_timezone = self._timezone
        c.author_timezone = self._timezone
        c.message = message.encode(encoding)
        if pseudoheaders:
            if not c.message.endswith(b"\n"):
                c.message += b"\n"
            c.message += b"\n" + b"".join([line + b"\n" for line in pseudoheaders])
        create_signatures = self._config_stack.get('create_signatures')
        if (create_signatures in (
                _mod_config.SIGN_ALWAYS, _mod_config.SIGN_WHEN_POSSIBLE)):
            strategy = gpg.GPGStrategy(self._config_stack)
            try:
                c.gpgsig = strategy.sign(c.as_raw_string(), gpg.MODE_DETACH)
            except gpg.GpgNotInstalled as e:
                if create_signatures == _mod_config.SIGN_WHEN_POSSIBLE:
                    trace.note('skipping commit signature: %s', e)
                else:
                    raise
            except gpg.SigningFailed as e:
                if create_signatures == _mod_config.SIGN_WHEN_POSSIBLE:
                    trace.note('commit signature failed: %s', e)
                else:
                    raise
        self.store.add_object(c)
        self.repository.commit_write_group()
        self._new_revision_id = self._mapping.revision_id_foreign_to_bzr(c.id)
        return self._new_revision_id

    def abort(self):
        if self.repository.is_in_write_group():
            self.repository.abort_write_group()

    def revision_tree(self):
        return self.repository.revision_tree(self._new_revision_id)

    def get_basis_delta(self):
        return self._inv_delta

    def update_basis_by_delta(self, revid, delta):
        pass



# Copyright (C) 2005-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

# The newly committed revision is going to have a shape corresponding
# to that of the working tree.  Files that are not in the
# working tree and that were in the predecessor are reported as
# removed --- this can include files that were either removed from the
# inventory or deleted in the working tree.  If they were only
# deleted from disk, they are removed from the working inventory.

# We then consider the remaining entries, which will be in the new
# version.  Directory entries are simply copied across.  File entries
# must be checked to see if a new version of the file should be
# recorded.  For each parent revision tree, we check to see what
# version of the file was present.  If the file was present in at
# least one tree, and if it was the same version in all the trees,
# then we can just refer to that version.  Otherwise, a new version
# representing the merger of the file versions must be added.

# TODO: Update hashcache before and after - or does the WorkingTree
# look after that?

# TODO: Rather than mashing together the ancestry and storing it back,
# perhaps the weave should have single method which does it all in one
# go, avoiding a lot of redundant work.

# TODO: Perhaps give a warning if one of the revisions marked as
# merged is already in the ancestry, and then don't record it as a
# distinct parent.

# TODO: If the file is newly merged but unchanged from the version it
# merges from, then it should still be reported as newly added
# relative to the basis revision.

# TODO: Change the parameter 'rev_id' to 'revision_id' to be consistent with
# the rest of the code; add a deprecation of the old name.

from contextlib import ExitStack

import breezy.config

from . import debug, errors, trace, ui
from .branch import Branch
from .errors import BzrError, ConflictsInTree, StrictCommitFailed
from .i18n import gettext
from .osutils import (get_user_encoding, is_inside_any, minimum_path_selection,
                      supports_symlinks)
from .trace import is_quiet, mutter, note
from .urlutils import unescape_for_display


class PointlessCommit(BzrError):

    _fmt = "No changes to commit"


class CannotCommitSelectedFileMerge(BzrError):

    _fmt = 'Selected-file commit of merges is not supported yet:'\
        ' files %(files_str)s'

    def __init__(self, files):
        files_str = ', '.join(files)
        BzrError.__init__(self, files=files, files_str=files_str)


def filter_excluded(iter_changes, exclude):
    """Filter exclude filenames.

    :param iter_changes: iter_changes function
    :param exclude: List of paths to exclude
    :return: iter_changes function
    """
    for change in iter_changes:
        new_excluded = (change.path[1] is not None and
                        is_inside_any(exclude, change.path[1]))

        old_excluded = (change.path[0] is not None and
                        is_inside_any(exclude, change.path[0]))

        if old_excluded and new_excluded:
            continue

        if old_excluded or new_excluded:
            # TODO(jelmer): Perhaps raise an error here instead?
            continue

        yield change


class NullCommitReporter:
    """I report on progress of a commit."""

    def started(self, revno, revid, location):
        pass

    def snapshot_change(self, change, path):
        pass

    def completed(self, revno, rev_id):
        pass

    def deleted(self, path):
        pass

    def missing(self, path):
        pass

    def renamed(self, change, old_path, new_path):
        pass

    def is_verbose(self):
        return False


class ReportCommitToLog(NullCommitReporter):

    def _note(self, format, *args):
        """Output a message.

        Subclasses may choose to override this method.
        """
        note(format, *args)

    def snapshot_change(self, change, path):
        if path == '' and change in (gettext('added'), gettext('modified')):
            return
        self._note("%s %s", change, path)

    def started(self, revno, rev_id, location):
        self._note(
            gettext('Committing to: %s'),
            unescape_for_display(location, 'utf-8'))

    def completed(self, revno, rev_id):
        if revno is not None:
            self._note(gettext('Committed revision %d.'), revno)
            # self._note goes to the console too; so while we want to log the
            # rev_id, we can't trivially only log it. (See bug 526425). Long
            # term we should rearrange the reporting structure, but for now
            # we just mutter seperately. We mutter the revid and revno together
            # so that concurrent bzr invocations won't lead to confusion.
            mutter('Committed revid %s as revno %d.', rev_id, revno)
        else:
            self._note(gettext('Committed revid %s.'), rev_id)

    def deleted(self, path):
        self._note(gettext('deleted %s'), path)

    def missing(self, path):
        self._note(gettext('missing %s'), path)

    def renamed(self, change, old_path, new_path):
        self._note('%s %s => %s', change, old_path, new_path)

    def is_verbose(self):
        return True


class Commit:
    """Task of committing a new revision.

    This is a MethodObject: it accumulates state as the commit is
    prepared, and then it is discarded.  It doesn't represent
    historical revisions, just the act of recording a new one.

            missing_ids
            Modified to hold a list of files that have been deleted from
            the working directory; these should be removed from the
            working inventory.
    """

    def __init__(self,
                 reporter=None,
                 config_stack=None):
        """Create a Commit object.

        :param reporter: the default reporter to use or None to decide later
        """
        self.reporter = reporter
        self.config_stack = config_stack

    @staticmethod
    def update_revprops(revprops, branch, authors=None,
                        local=False, possible_master_transports=None):
        if revprops is None:
            revprops = {}
        if possible_master_transports is None:
            possible_master_transports = []
        if ('branch-nick' not in revprops and
                branch.repository._format.supports_storing_branch_nick):
            revprops['branch-nick'] = branch._get_nick(
                local,
                possible_master_transports)
        if authors is not None:
            if 'author' in revprops or 'authors' in revprops:
                # XXX: maybe we should just accept one of them?
                raise AssertionError('author property given twice')
            if authors:
                for individual in authors:
                    if '\n' in individual:
                        raise AssertionError('\\n is not a valid character '
                                             'in an author identity')
                revprops['authors'] = '\n'.join(authors)
        return revprops

    def commit(self,
               message=None,
               timestamp=None,
               timezone=None,
               committer=None,
               specific_files=None,
               rev_id=None,
               allow_pointless=True,
               strict=False,
               verbose=False,
               revprops=None,
               working_tree=None,
               local=False,
               reporter=None,
               config=None,
               message_callback=None,
               recursive='down',
               exclude=None,
               possible_master_transports=None,
               lossy=False):
        """Commit working copy as a new revision.

        :param message: the commit message (it or message_callback is required)
        :param message_callback: A callback: message =
            message_callback(cmt_obj)

        :param timestamp: if not None, seconds-since-epoch for a
            postdated/predated commit.

        :param specific_files: If not None, commit only those files. An empty
            list means 'commit no files'.

        :param rev_id: If set, use this as the new revision id.
            Useful for test or import commands that need to tightly
            control what revisions are assigned.  If you duplicate
            a revision id that exists elsewhere it is your own fault.
            If null (default), a time/random revision id is generated.

        :param allow_pointless: If true (default), commit even if nothing
            has changed and no merges are recorded.

        :param strict: If true, don't allow a commit if the working tree
            contains unknown files.

        :param revprops: Properties for new revision
        :param local: Perform a local only commit.
        :param reporter: the reporter to use or None for the default
        :param verbose: if True and the reporter is not None, report everything
        :param recursive: If set to 'down', commit in any subtrees that have
            pending changes of any sort during this commit.
        :param exclude: None or a list of relative paths to exclude from the
            commit. Pending changes to excluded files will be ignored by the
            commit.
        :param lossy: When committing to a foreign VCS, ignore any
            data that can not be natively represented.
        """
        with ExitStack() as stack:
            self.revprops = revprops or {}
            # XXX: Can be set on __init__ or passed in - this is a bit ugly.
            self.config_stack = config or self.config_stack
            mutter('preparing to commit')

            if working_tree is None:
                raise BzrError("working_tree must be passed into commit().")
            else:
                self.work_tree = working_tree
                self.branch = self.work_tree.branch
                if getattr(self.work_tree, 'requires_rich_root', lambda: False)():
                    if not self.branch.repository.supports_rich_root():
                        raise errors.RootNotRich()
            if message_callback is None:
                if message is not None:
                    if isinstance(message, bytes):
                        message = message.decode(get_user_encoding())

                    def message_callback(x):
                        return message
                else:
                    raise BzrError("The message or message_callback keyword"
                                   " parameter is required for commit().")

            self.bound_branch = None
            self.any_entries_deleted = False
            if exclude is not None:
                self.exclude = sorted(
                    minimum_path_selection(exclude))
            else:
                self.exclude = []
            self.local = local
            self.master_branch = None
            self.recursive = recursive
            self.rev_id = None
            # self.specific_files is None to indicate no filter, or any iterable to
            # indicate a filter - [] means no files at all, as per iter_changes.
            if specific_files is not None:
                self.specific_files = sorted(
                    minimum_path_selection(specific_files))
            else:
                self.specific_files = None

            self.allow_pointless = allow_pointless
            self.message_callback = message_callback
            self.timestamp = timestamp
            self.timezone = timezone
            self.committer = committer
            self.strict = strict
            self.verbose = verbose

            stack.enter_context(self.work_tree.lock_write())
            self.parents = self.work_tree.get_parent_ids()
            self.pb = ui.ui_factory.nested_progress_bar()
            stack.callback(self.pb.finished)
            self.basis_revid = self.work_tree.last_revision()
            self.basis_tree = self.work_tree.basis_tree()
            stack.enter_context(self.basis_tree.lock_read())
            # Cannot commit with conflicts present.
            if len(self.work_tree.conflicts()) > 0:
                raise ConflictsInTree

            # Setup the bound branch variables as needed.
            self._check_bound_branch(stack, possible_master_transports)
            if self.config_stack is None:
                self.config_stack = self.work_tree.get_config_stack()

            # Check that the working tree is up to date
            old_revno, old_revid, new_revno = self._check_out_of_date_tree()

            # Complete configuration setup
            if reporter is not None:
                self.reporter = reporter
            elif self.reporter is None:
                self.reporter = self._select_reporter()

            # Setup the progress bar. As the number of files that need to be
            # committed in unknown, progress is reported as stages.
            # We keep track of entries separately though and include that
            # information in the progress bar during the relevant stages.
            self.pb_stage_name = ""
            self.pb_stage_count = 0
            self.pb_stage_total = 5
            if self.bound_branch:
                # 2 extra stages: "Uploading data to master branch" and "Merging
                # tags to master branch"
                self.pb_stage_total += 2
            self.pb.show_pct = False
            self.pb.show_spinner = False
            self.pb.show_eta = False
            self.pb.show_count = True
            self.pb.show_bar = True

            # After a merge, a selected file commit is not supported.
            # See 'bzr help merge' for an explanation as to why.
            if len(self.parents) > 1 and self.specific_files is not None:
                raise CannotCommitSelectedFileMerge(self.specific_files)
            # Excludes are a form of selected file commit.
            if len(self.parents) > 1 and self.exclude:
                raise CannotCommitSelectedFileMerge(self.exclude)

            # Collect the changes
            self._set_progress_stage("Collecting changes", counter=True)
            self._lossy = lossy
            self.builder = self.branch.get_commit_builder(
                self.parents, self.config_stack, timestamp, timezone, committer,
                self.revprops, rev_id, lossy=lossy)

            if self.builder.updates_branch and self.bound_branch:
                self.builder.abort()
                raise AssertionError(
                    "bound branches not supported for commit builders "
                    "that update the branch")

            try:
                # find the location being committed to
                if self.bound_branch:
                    master_location = self.master_branch.base
                else:
                    master_location = self.branch.base

                # report the start of the commit
                self.reporter.started(new_revno, self.rev_id, master_location)

                self._update_builder_with_changes()
                self._check_pointless()

                # TODO: Now the new inventory is known, check for conflicts.
                # ADHB 2006-08-08: If this is done, populate_new_inv should not add
                # weave lines, because nothing should be recorded until it is known
                # that commit will succeed.
                self._set_progress_stage("Saving data locally")
                self.builder.finish_inventory()

                # Prompt the user for a commit message if none provided
                message = message_callback(self)
                self.message = message

                # Add revision data to the local branch
                self.rev_id = self.builder.commit(self.message)

            except Exception:
                mutter("aborting commit write group because of exception:")
                trace.log_exception_quietly()
                self.builder.abort()
                raise

            self._update_branches(old_revno, old_revid, new_revno)

            # Make the working tree be up to date with the branch. This
            # includes automatic changes scheduled to be made to the tree, such
            # as updating its basis and unversioning paths that were missing.
            self.work_tree.unversion(self.deleted_paths)
            self._set_progress_stage("Updating the working tree")
            self.work_tree.update_basis_by_delta(
                self.rev_id, self.builder.get_basis_delta())
            self.reporter.completed(new_revno, self.rev_id)
            self._process_post_hooks(old_revno, new_revno)
            return self.rev_id

    def _update_branches(self, old_revno, old_revid, new_revno):
        """Update the master and local branch to the new revision.

        This will try to make sure that the master branch is updated
        before the local branch.

        :param old_revno: Revision number of master branch before the
            commit
        :param old_revid: Tip of master branch before the commit
        :param new_revno: Revision number of the new commit
        """
        if not self.builder.updates_branch:
            self._process_pre_hooks(old_revno, new_revno)

            # Upload revision data to the master.
            # this will propagate merged revisions too if needed.
            if self.bound_branch:
                self._set_progress_stage("Uploading data to master branch")
                # 'commit' to the master first so a timeout here causes the
                # local branch to be out of date
                (new_revno, self.rev_id) = self.master_branch.import_last_revision_info_and_tags(
                    self.branch, new_revno, self.rev_id, lossy=self._lossy)
                if self._lossy:
                    self.branch.fetch(self.master_branch, self.rev_id)

            # and now do the commit locally.
            if new_revno is None:
                # Keep existing behaviour around ghosts
                new_revno = 1
            self.branch.set_last_revision_info(new_revno, self.rev_id)
        else:
            try:
                self._process_pre_hooks(old_revno, new_revno)
            except BaseException:
                # The commit builder will already have updated the branch,
                # revert it.
                self.branch.set_last_revision_info(old_revno, old_revid)
                raise

        # Merge local tags to remote
        if self.bound_branch:
            self._set_progress_stage("Merging tags to master branch")
            tag_updates, tag_conflicts = self.branch.tags.merge_to(
                self.master_branch.tags)
            if tag_conflicts:
                warning_lines = ['    ' + name for name, _, _ in tag_conflicts]
                note(gettext("Conflicting tags in bound branch:\n{}".format(
                    "\n".join(warning_lines))))

    def _select_reporter(self):
        """Select the CommitReporter to use."""
        if is_quiet():
            return NullCommitReporter()
        return ReportCommitToLog()

    def _check_pointless(self):
        if self.allow_pointless:
            return
        # A merge with no effect on files
        if len(self.parents) > 1:
            return
        if self.builder.any_changes():
            return
        raise PointlessCommit()

    def _check_bound_branch(self, stack, possible_master_transports=None):
        """Check to see if the local branch is bound.

        If it is bound, then most of the commit will actually be
        done using the remote branch as the target branch.
        Only at the end will the local branch be updated.
        """
        if self.local and not self.branch.get_bound_location():
            raise errors.LocalRequiresBoundBranch()

        if not self.local:
            self.master_branch = self.branch.get_master_branch(
                possible_master_transports)

        if not self.master_branch:
            # make this branch the reference branch for out of date checks.
            self.master_branch = self.branch
            return

        # If the master branch is bound, we must fail
        master_bound_location = self.master_branch.get_bound_location()
        if master_bound_location:
            raise errors.CommitToDoubleBoundBranch(
                self.branch, self.master_branch, master_bound_location)

        # TODO: jam 20051230 We could automatically push local
        #       commits to the remote branch if they would fit.
        #       But for now, just require remote to be identical
        #       to local.

        # Make sure the local branch is identical to the master
        master_revid = self.master_branch.last_revision()
        local_revid = self.branch.last_revision()
        if local_revid != master_revid:
            raise errors.BoundBranchOutOfDate(self.branch,
                                              self.master_branch)

        # Now things are ready to change the master branch
        # so grab the lock
        self.bound_branch = self.branch
        stack.enter_context(self.master_branch.lock_write())

    def _check_out_of_date_tree(self):
        """Check that the working tree is up to date.

        :return: old_revision_number, old_revision_id, new_revision_number
            tuple
        """
        try:
            first_tree_parent = self.work_tree.get_parent_ids()[0]
        except IndexError:
            # if there are no parents, treat our parent as 'None'
            # this is so that we still consider the master branch
            # - in a checkout scenario the tree may have no
            # parents but the branch may do.
            first_tree_parent = breezy.revision.NULL_REVISION
        if (self.master_branch._format.stores_revno() or
                self.config_stack.get('calculate_revnos')):
            try:
                old_revno, master_last = self.master_branch.last_revision_info()
            except errors.UnsupportedOperation:
                master_last = self.master_branch.last_revision()
                old_revno = self.branch.revision_id_to_revno(master_last)
        else:
            master_last = self.master_branch.last_revision()
            old_revno = None
        if master_last != first_tree_parent:
            if master_last != breezy.revision.NULL_REVISION:
                raise errors.OutOfDateTree(self.work_tree)
        if (old_revno is not None and
                self.branch.repository.has_revision(first_tree_parent)):
            new_revno = old_revno + 1
        else:
            # ghost parents never appear in revision history.
            new_revno = None
        return old_revno, master_last, new_revno

    def _process_pre_hooks(self, old_revno, new_revno):
        """Process any registered pre commit hooks."""
        self._set_progress_stage("Running pre_commit hooks")
        self._process_hooks("pre_commit", old_revno, new_revno)

    def _process_post_hooks(self, old_revno, new_revno):
        """Process any registered post commit hooks."""
        # Process the post commit hooks, if any
        self._set_progress_stage("Running post_commit hooks")
        # old style commit hooks - should be deprecated ? (obsoleted in
        # 0.15^H^H^H^H 2.5.0)
        post_commit = self.config_stack.get('post_commit')
        if post_commit is not None:
            hooks = post_commit.split(' ')
            # this would be nicer with twisted.python.reflect.namedAny
            for hook in hooks:
                result = eval(hook + '(branch, rev_id)',
                              {'branch': self.branch,
                               'breezy': breezy,
                               'rev_id': self.rev_id})
        # process new style post commit hooks
        self._process_hooks("post_commit", old_revno, new_revno)

    def _process_hooks(self, hook_name, old_revno, new_revno):
        if not Branch.hooks[hook_name]:
            return

        # new style commit hooks:
        if not self.bound_branch:
            hook_master = self.branch
            hook_local = None
        else:
            hook_master = self.master_branch
            hook_local = self.branch
        # With bound branches, when the master is behind the local branch,
        # the 'old_revno' and old_revid values here are incorrect.
        # XXX: FIXME ^. RBC 20060206
        if self.parents:
            old_revid = self.parents[0]
        else:
            old_revid = breezy.revision.NULL_REVISION

        if hook_name == "pre_commit":
            future_tree = self.builder.revision_tree()
            tree_delta = future_tree.changes_from(self.basis_tree,
                                                  include_root=True)

        for hook in Branch.hooks[hook_name]:
            # show the running hook in the progress bar. As hooks may
            # end up doing nothing (e.g. because they are not configured by
            # the user) this is still showing progress, not showing overall
            # actions - its up to each plugin to show a UI if it want's to
            # (such as 'Emailing diff to foo@example.com').
            self.pb_stage_name = "Running %s hooks [%s]" % \
                (hook_name, Branch.hooks.get_hook_name(hook))
            self._emit_progress()
            if 'hooks' in debug.debug_flags:
                mutter("Invoking commit hook: %r", hook)
            if hook_name == "post_commit":
                hook(hook_local, hook_master, old_revno, old_revid, new_revno,
                     self.rev_id)
            elif hook_name == "pre_commit":
                hook(hook_local, hook_master,
                     old_revno, old_revid, new_revno, self.rev_id,
                     tree_delta, future_tree)

    def _update_builder_with_changes(self):
        """Update the commit builder with the data about what has changed.
        """
        specific_files = self.specific_files
        mutter("Selecting files for commit with filter %r", specific_files)

        self._check_strict()
        iter_changes = self.work_tree.iter_changes(
            self.basis_tree, specific_files=specific_files)
        if self.exclude:
            iter_changes = filter_excluded(iter_changes, self.exclude)
        iter_changes = self._filter_iter_changes(iter_changes)
        for path, fs_hash in self.builder.record_iter_changes(
                self.work_tree, self.basis_revid, iter_changes):
            self.work_tree._observed_sha1(path, fs_hash)

    def _filter_iter_changes(self, iter_changes):
        """Process iter_changes.

        This method reports on the changes in iter_changes to the user, and
        converts 'missing' entries in the iter_changes iterator to 'deleted'
        entries. 'missing' entries have their

        :param iter_changes: An iter_changes to process.
        :return: A generator of changes.
        """
        reporter = self.reporter
        report_changes = reporter.is_verbose()
        deleted_paths = []
        for change in iter_changes:
            if report_changes:
                old_path = change.path[0]
                new_path = change.path[1]
                versioned = change.versioned[1]
            kind = change.kind[1]
            versioned = change.versioned[1]
            if kind is None and versioned:
                # 'missing' path
                if report_changes:
                    reporter.missing(new_path)
                if change.kind[0] == 'symlink' and not self.work_tree.supports_symlinks():
                    trace.warning('Ignoring "%s" as symlinks are not '
                                  'supported on this filesystem.' % (change.path[0],))
                    continue
                deleted_paths.append(change.path[1])
                # Reset the new path (None) and new versioned flag (False)
                change = change.discard_new()
                new_path = change.path[1]
                versioned = False
            elif kind == 'tree-reference':
                if self.recursive == 'down':
                    self._commit_nested_tree(change.path[1])
            if change.versioned[0] or change.versioned[1]:
                yield change
                if report_changes:
                    if new_path is None:
                        reporter.deleted(old_path)
                    elif old_path is None:
                        reporter.snapshot_change(gettext('added'), new_path)
                    elif old_path != new_path:
                        reporter.renamed(gettext('renamed'),
                                         old_path, new_path)
                    else:
                        if (new_path
                                or self.work_tree.branch.repository._format.rich_root_data):
                            # Don't report on changes to '' in non rich root
                            # repositories.
                            reporter.snapshot_change(
                                gettext('modified'), new_path)
            self._next_progress_entry()
        # Unversion files that were found to be deleted
        self.deleted_paths = deleted_paths

    def _check_strict(self):
        # XXX: when we use iter_changes this would likely be faster if
        # iter_changes would check for us (even in the presence of
        # selected_files).
        if self.strict:
            # raise an exception as soon as we find a single unknown.
            for unknown in self.work_tree.unknowns():
                raise StrictCommitFailed()

    def _commit_nested_tree(self, path):
        "Commit a nested tree."
        sub_tree = self.work_tree.get_nested_tree(path)
        # FIXME: be more comprehensive here:
        # this works when both trees are in --trees repository,
        # but when both are bound to a different repository,
        # it fails; a better way of approaching this is to
        # finally implement the explicit-caches approach design
        # a while back - RBC 20070306.
        if sub_tree.branch.repository.has_same_location(
                self.work_tree.branch.repository):
            sub_tree.branch.repository = \
                self.work_tree.branch.repository
        try:
            return sub_tree.commit(message=None, revprops=self.revprops,
                                   recursive=self.recursive,
                                   message_callback=self.message_callback,
                                   timestamp=self.timestamp,
                                   timezone=self.timezone,
                                   committer=self.committer,
                                   allow_pointless=self.allow_pointless,
                                   strict=self.strict, verbose=self.verbose,
                                   local=self.local, reporter=self.reporter)
        except PointlessCommit:
            return self.work_tree.get_reference_revision(path)

    def _set_progress_stage(self, name, counter=False):
        """Set the progress stage and emit an update to the progress bar."""
        self.pb_stage_name = name
        self.pb_stage_count += 1
        if counter:
            self.pb_entries_count = 0
        else:
            self.pb_entries_count = None
        self._emit_progress()

    def _next_progress_entry(self):
        """Emit an update to the progress bar and increment the entry count."""
        self.pb_entries_count += 1
        self._emit_progress()

    def _emit_progress(self):
        if self.pb_entries_count is not None:
            text = gettext("{0} [{1}] - Stage").format(self.pb_stage_name,
                                                       self.pb_entries_count)
        else:
            text = gettext("%s - Stage") % (self.pb_stage_name, )
        self.pb.update(text, self.pb_stage_count, self.pb_stage_total)



# Copyright (C) 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Transport implementation that doesn't detect clashing renames.
"""

from .. import urlutils
from ..transport import FileExists
from . import decorator


class BrokenRenameTransportDecorator(decorator.TransportDecorator):
    """A transport that fails to detect clashing renames"""

    @classmethod
    def _get_url_prefix(self):
        """FakeNFS transports are identified by 'brokenrename+'"""
        return 'brokenrename+'

    def rename(self, rel_from, rel_to):
        """See Transport.rename().
        """
        try:
            if self._decorated.has(rel_to):
                rel_to = urlutils.join(rel_to, urlutils.basename(rel_from))
            self._decorated.rename(rel_from, rel_to)
        except (errors.DirectoryNotEmpty, FileExists) as e:
            # absorb the error
            return


def get_test_permutations():
    """Return the permutations to be used in testing."""
    # we don't use this for general testing, only for the tests that
    # specifically want it
    return []



# Copyright (C) 2020 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Tests for breezy.git.tests."""

import stat

from dulwich.objects import Blob, Tree

from breezy.bzr.inventorytree import InventoryTreeChange as TreeChange
from breezy.delta import TreeDelta
from breezy.errors import PathsNotVersionedError
from breezy.git.mapping import default_mapping
from breezy.git.tree import (changes_from_git_changes,
                             tree_delta_from_git_changes)
from breezy.tests import TestCase, TestCaseWithTransport
from breezy.workingtree import WorkingTree

REG_MODE = stat.S_IFREG | 0o644


class ChangesFromGitChangesTests(TestCase):

    def setUp(self):
        super().setUp()
        self.maxDiff = None
        self.mapping = default_mapping

    def transform(
            self, changes, specific_files=None, include_unchanged=False,
            source_extras=None, target_extras=None):
        return list(changes_from_git_changes(
            changes, self.mapping, specific_files=specific_files,
            include_unchanged=include_unchanged, source_extras=source_extras,
            target_extras=target_extras))

    def test_empty(self):
        self.assertEqual([], self.transform([]))

    def test_modified(self):
        a = Blob.from_string(b'a')
        b = Blob.from_string(b'b')
        self.assertEqual([
            TreeChange(
                b'git:a', ('a', 'a'), True, (True, True),
                (b'TREE_ROOT', b'TREE_ROOT'), ('a', 'a'), ('file', 'file'),
                (False, False), False)
            ], self.transform([
                ('modify',
                 (b'a', stat.S_IFREG | 0o644, a), (b'a', stat.S_IFREG | 0o644, b))]))

    def test_kind_changed(self):
        a = Blob.from_string(b'a')
        b = Blob.from_string(b'target')
        self.assertEqual([
            TreeChange(
                b'git:a', ('a', 'a'), True, (True, True),
                (b'TREE_ROOT', b'TREE_ROOT'), ('a', 'a'), ('file', 'symlink'),
                (False, False), False)
            ], self.transform([
                ('modify',
                 (b'a', stat.S_IFREG | 0o644, a), (b'a', stat.S_IFLNK, b))]))

    def test_rename_no_changes(self):
        a = Blob.from_string(b'a')
        self.assertEqual([
            TreeChange(
                b'git:old', ('old', 'a'), False, (True, True),
                (b'TREE_ROOT', b'TREE_ROOT'), ('old', 'a'),
                ('file', 'file'), (False, False), False)
            ], self.transform([
                ('rename',
                 (b'old', stat.S_IFREG | 0o644, a), (b'a', stat.S_IFREG | 0o644, a))]))

    def test_rename_and_modify(self):
        a = Blob.from_string(b'a')
        b = Blob.from_string(b'b')
        self.assertEqual([
            TreeChange(
                b'git:a', ('a', 'b'), True, (True, True),
                (b'TREE_ROOT', b'TREE_ROOT'), ('a', 'b'),
                ('file', 'file'), (False, False), False)
            ], self.transform([
                ('rename',
                 (b'a', stat.S_IFREG | 0o644, a), (b'b', stat.S_IFREG | 0o644, b))]))

    def test_copy_no_changes(self):
        a = Blob.from_string(b'a')
        self.assertEqual([
            TreeChange(
                b'git:a', ('old', 'a'), False, (True, True),
                (b'TREE_ROOT', b'TREE_ROOT'), ('old', 'a'),
                ('file', 'file'), (False, False), True)
            ], self.transform([
                ('copy',
                 (b'old', stat.S_IFREG | 0o644, a), (b'a', stat.S_IFREG | 0o644, a))]))

    def test_copy_and_modify(self):
        a = Blob.from_string(b'a')
        b = Blob.from_string(b'b')
        self.assertEqual([
            TreeChange(
                b'git:b', ('a', 'b'), True, (True, True),
                (b'TREE_ROOT', b'TREE_ROOT'), ('a', 'b'),
                ('file', 'file'), (False, False), True)
            ], self.transform([
                ('copy',
                 (b'a', stat.S_IFREG | 0o644, a), (b'b', stat.S_IFREG | 0o644, b))]))

    def test_add(self):
        b = Blob.from_string(b'b')
        self.assertEqual([
            TreeChange(
                b'git:a', (None, 'a'), True, (False, True), (None, b'TREE_ROOT'),
                (None, 'a'), (None, 'file'), (None, False), False)
            ], self.transform([
                ('add',
                 (None, None, None), (b'a', stat.S_IFREG | 0o644, b))]))

    def test_delete(self):
        b = Blob.from_string(b'b')
        self.assertEqual([
            TreeChange(
                b'git:a', ('a', None), True, (True, False), (b'TREE_ROOT', None),
                ('a', None), ('file', None), (False, None), False)
            ], self.transform([
                ('remove',
                 (b'a', stat.S_IFREG | 0o644, b), (None, None, None))]))

    def test_unchanged(self):
        b = Blob.from_string(b'b')
        self.assertEqual([
            TreeChange(
                b'git:a', ('a', 'a'), False, (True, True),
                (b'TREE_ROOT', b'TREE_ROOT'), ('a', 'a'), ('file', 'file'),
                (False, False), False)
            ], self.transform([
                ('unchanged',
                 (b'a', stat.S_IFREG | 0o644, b), (b'a', stat.S_IFREG | 0o644, b))],
            include_unchanged=True))
        self.assertEqual([], self.transform([
            ('unchanged',
             (b'a', stat.S_IFREG | 0o644, b), (b'a', stat.S_IFREG | 0o644, b))],
            include_unchanged=False))

    def test_unversioned(self):
        b = Blob.from_string(b'b')
        self.assertEqual([
            TreeChange(
                None, (None, 'a'), True, (False, False),
                (None, b'TREE_ROOT'), (None, 'a'), (None, 'file'),
                (None, False), False)
            ], self.transform([
                ('add',
                 (None, None, None), (b'a', stat.S_IFREG | 0o644, b))],
            target_extras={b'a'}))
        self.assertEqual([
            TreeChange(
                None, ('a', 'a'), False, (False, False),
                (b'TREE_ROOT', b'TREE_ROOT'), ('a', 'a'), ('file', 'file'),
                (False, False), False)
            ], self.transform([
                ('add',
                 (b'a', stat.S_IFREG | 0o644, b), (b'a', stat.S_IFREG | 0o644, b))],
            source_extras={b'a'},
            target_extras={b'a'}))


class DeltaFromGitChangesTests(TestCase):

    def setUp(self):
        super().setUp()
        self.maxDiff = None
        self.mapping = default_mapping

    def transform(
            self, changes, specific_files=None,
            require_versioned=False, include_root=False,
            source_extras=None, target_extras=None):
        return tree_delta_from_git_changes(
            changes, (self.mapping, self.mapping),
            specific_files=specific_files,
            require_versioned=require_versioned, include_root=include_root,
            source_extras=source_extras, target_extras=target_extras)

    def test_empty(self):
        self.assertEqual(TreeDelta(), self.transform([]))

    def test_modified(self):
        a = Blob.from_string(b'a')
        b = Blob.from_string(b'b')
        delta = self.transform([
            ('modify',
             (b'a', stat.S_IFREG | 0o644, a), (b'a', stat.S_IFREG | 0o644, b))])
        expected_delta = TreeDelta()
        expected_delta.modified.append(TreeChange(
            b'git:a', ('a', 'a'), True, (True, True),
            (b'TREE_ROOT', b'TREE_ROOT'), ('a', 'a'), ('file', 'file'),
            (False, False), False))
        self.assertEqual(expected_delta, delta)

    def test_rename_no_changes(self):
        a = Blob.from_string(b'a')
        delta = self.transform([
            ('rename',
             (b'old', stat.S_IFREG | 0o644, a), (b'a', stat.S_IFREG | 0o644, a))])
        expected_delta = TreeDelta()
        expected_delta.renamed.append(
            TreeChange(
                b'git:old', ('old', 'a'), False, (True, True),
                (b'TREE_ROOT', b'TREE_ROOT'), ('old', 'a'),
                ('file', 'file'), (False, False), False))
        self.assertEqual(expected_delta, delta)

    def test_rename_and_modify(self):
        a = Blob.from_string(b'a')
        b = Blob.from_string(b'b')
        delta = self.transform([
            ('rename',
             (b'a', stat.S_IFREG | 0o644, a), (b'b', stat.S_IFREG | 0o644, b))])
        expected_delta = TreeDelta()
        expected_delta.renamed.append(
            TreeChange(
                b'git:a', ('a', 'b'), True, (True, True),
                (b'TREE_ROOT', b'TREE_ROOT'), ('a', 'b'),
                ('file', 'file'), (False, False), False))
        self.assertEqual(delta, expected_delta)

    def test_copy_no_changes(self):
        a = Blob.from_string(b'a')
        delta = self.transform([
            ('copy',
             (b'old', stat.S_IFREG | 0o644, a), (b'a', stat.S_IFREG | 0o644, a))])
        expected_delta = TreeDelta()
        expected_delta.copied.append(TreeChange(
            b'git:a', ('old', 'a'), False, (True, True),
            (b'TREE_ROOT', b'TREE_ROOT'), ('old', 'a'),
            ('file', 'file'), (False, False), True))
        self.assertEqual(expected_delta, delta)

    def test_copy_and_modify(self):
        a = Blob.from_string(b'a')
        b = Blob.from_string(b'b')
        delta = self.transform([
            ('copy',
             (b'a', stat.S_IFREG | 0o644, a),
             (b'b', stat.S_IFREG | 0o644, b))])
        expected_delta = TreeDelta()
        expected_delta.copied.append(TreeChange(
            b'git:b', ('a', 'b'), True, (True, True),
            (b'TREE_ROOT', b'TREE_ROOT'), ('a', 'b'),
            ('file', 'file'), (False, False), True))
        self.assertEqual(expected_delta, delta)

    def test_add(self):
        b = Blob.from_string(b'b')
        delta = self.transform([
            ('add',
             (None, None, None), (b'a', stat.S_IFREG | 0o644, b))])
        expected_delta = TreeDelta()
        expected_delta.added.append(TreeChange(
            b'git:a', (None, 'a'), True, (False, True), (None, b'TREE_ROOT'),
            (None, 'a'), (None, 'file'), (None, False), False))
        self.assertEqual(delta, expected_delta)

    def test_delete(self):
        b = Blob.from_string(b'b')
        delta = self.transform([
            ('remove',
             (b'a', stat.S_IFREG | 0o644, b), (None, None, None))])
        expected_delta = TreeDelta()
        expected_delta.removed.append(TreeChange(
            b'git:a', ('a', None), True, (True, False), (b'TREE_ROOT', None),
            ('a', None), ('file', None), (False, None), False))
        self.assertEqual(delta, expected_delta)

    def test_unchanged(self):
        b = Blob.from_string(b'b')
        delta = self.transform([
            ('unchanged',
             (b'a', stat.S_IFREG | 0o644, b), (b'a', stat.S_IFREG | 0o644, b))])
        expected_delta = TreeDelta()
        expected_delta.unchanged.append(TreeChange(
            b'git:a', ('a', 'a'), False, (True, True),
            (b'TREE_ROOT', b'TREE_ROOT'), ('a', 'a'), ('file', 'file'),
            (False, False), False))

    def test_unversioned(self):
        b = Blob.from_string(b'b')
        delta = self.transform([
            ('add',
             (None, None, None), (b'a', stat.S_IFREG | 0o644, b))],
            target_extras={b'a'})
        expected_delta = TreeDelta()
        expected_delta.unversioned.append(
            TreeChange(
                None, (None, 'a'), True, (False, False),
                (None, b'TREE_ROOT'), (None, 'a'), (None, 'file'),
                (None, False), False))
        self.assertEqual(delta, expected_delta)
        delta = self.transform([
            ('add',
             (b'a', stat.S_IFREG | 0o644, b), (b'a', stat.S_IFREG | 0o644, b))],
            source_extras={b'a'},
            target_extras={b'a'})
        expected_delta = TreeDelta()
        expected_delta.unversioned.append(TreeChange(
            None, ('a', 'a'), False, (False, False),
            (b'TREE_ROOT', b'TREE_ROOT'), ('a', 'a'), ('file', 'file'),
            (False, False), False))
        self.assertEqual(delta, expected_delta)

    def test_kind_change(self):
        a = Blob.from_string(b'a')
        b = Blob.from_string(b'target')
        delta = self.transform([
            ('modify',
             (b'a', stat.S_IFREG | 0o644, a), (b'a', stat.S_IFLNK, b))])
        expected_delta = TreeDelta()
        expected_delta.kind_changed.append(TreeChange(
            b'git:a', ('a', 'a'), True, (True, True),
            (b'TREE_ROOT', b'TREE_ROOT'), ('a', 'a'), ('file', 'symlink'),
            (False, False), False))
        self.assertEqual(expected_delta, delta)


class FindRelatedPathsAcrossTrees(TestCaseWithTransport):

    def test_none(self):
        self.make_branch_and_tree('t1', format='git')
        wt = WorkingTree.open('t1')
        self.assertIs(None, wt.find_related_paths_across_trees(None))

    def test_empty(self):
        self.make_branch_and_tree('t1', format='git')
        wt = WorkingTree.open('t1')
        self.assertEqual([], list(wt.find_related_paths_across_trees([])))

    def test_directory(self):
        self.make_branch_and_tree('t1', format='git')
        wt = WorkingTree.open('t1')
        self.build_tree(['t1/dir/', 't1/dir/file'])
        wt.add(['dir', 'dir/file'])
        self.assertEqual(['dir/file'], list(wt.find_related_paths_across_trees(['dir/file'])))
        self.assertEqual(['dir'], list(wt.find_related_paths_across_trees(['dir'])))

    def test_empty_directory(self):
        self.make_branch_and_tree('t1', format='git')
        wt = WorkingTree.open('t1')
        self.build_tree(['t1/dir/'])
        wt.add(['dir'])
        self.assertEqual(['dir'], list(wt.find_related_paths_across_trees(['dir'])))
        self.assertRaises(PathsNotVersionedError, wt.find_related_paths_across_trees, ['dir/file'])

    def test_missing(self):
        self.make_branch_and_tree('t1', format='git')
        wt = WorkingTree.open('t1')
        self.assertRaises(PathsNotVersionedError, wt.find_related_paths_across_trees, ['file'])

    def test_not_versioned(self):
        self.make_branch_and_tree('t1', format='git')
        self.make_branch_and_tree('t2', format='git')
        wt1 = WorkingTree.open('t1')
        wt2 = WorkingTree.open('t2')
        self.build_tree(['t1/file'])
        self.build_tree(['t2/file'])
        self.assertRaises(PathsNotVersionedError, wt1.find_related_paths_across_trees, ['file'], [wt2])

    def test_single(self):
        self.make_branch_and_tree('t1', format='git')
        wt = WorkingTree.open('t1')
        self.build_tree(['t1/file'])
        wt.add('file')
        self.assertEqual(['file'], list(wt.find_related_paths_across_trees(['file'])))



# Copyright (C) 2010 Canonical Ltd
# Authors: Robert Collins <robert.collins@canonical.com>
#          Jelmer Vernooij <jelmer.vernooij@canonical.com>
# -*- coding: utf-8 -*-
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""BzrDir implementation tests for bzr.

These test the conformance of all the bzrdir variations to the expected API.
Specific tests for individual formats are in the tests/test_bzrdir.py file
rather than in tests/per_branch/*.py. Generic control directory tests not
specific to BzrDir are in tests/per_controldir/*.py.
"""

from breezy.bzr.bzrdir import BzrDirFormat
from breezy.controldir import ControlDirFormat
from breezy.tests import (TestCaseWithTransport, default_transport,
                          multiply_tests, test_server)
from breezy.tests.per_controldir import make_scenarios
from breezy.transport import memory


class TestCaseWithBzrDir(TestCaseWithTransport):

    def setUp(self):
        super().setUp()
        self.controldir = None

    def get_bzrdir(self):
        if self.controldir is None:
            self.controldir = self.make_controldir(None)
        return self.controldir

    def get_default_format(self):
        return self.bzrdir_format


def load_tests(loader, standard_tests, pattern):
    test_per_bzrdir = [
        'breezy.bzr.tests.per_bzrdir.test_bzrdir',
        ]
    submod_tests = loader.loadTestsFromModuleNames(test_per_bzrdir)
    formats = [format for format in ControlDirFormat.known_formats()
               if isinstance(format, BzrDirFormat)]
    scenarios = make_scenarios(
        default_transport,
        None,
        # None here will cause a readonly decorator to be created
        # by the TestCaseWithTransport.get_readonly_transport method.
        None,
        formats)
    # This will always add scenarios using the smart server.
    from breezy.bzr.remote import RemoteBzrDirFormat

    # test the remote server behaviour when backed with a MemoryTransport
    # Once for the current version
    scenarios.extend(make_scenarios(
        memory.MemoryServer,
        test_server.SmartTCPServer_for_testing,
        test_server.ReadonlySmartTCPServer_for_testing,
        [(RemoteBzrDirFormat())],
        name_suffix='-default'))
    # And once with < 1.6 - the 'v2' protocol.
    scenarios.extend(make_scenarios(
        memory.MemoryServer,
        test_server.SmartTCPServer_for_testing_v2_only,
        test_server.ReadonlySmartTCPServer_for_testing_v2_only,
        [(RemoteBzrDirFormat())],
        name_suffix='-v2'))
    # add the tests for the sub modules
    return multiply_tests(submod_tests, scenarios, standard_tests)



# Copyright (C) 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for implementations of Repository.has_revisions."""

from breezy.revision import NULL_REVISION
from breezy.tests.per_repository import TestCaseWithRepository


class TestHasRevisions(TestCaseWithRepository):

    def test_empty_list(self):
        repo = self.make_repository('.')
        self.assertEqual(set(), repo.has_revisions([]))

    def test_superset(self):
        tree = self.make_branch_and_tree('.')
        repo = tree.branch.repository
        rev1 = tree.commit('1')
        rev2 = tree.commit('2')
        rev3 = tree.commit('3')
        self.assertEqual({rev1, rev3},
                         repo.has_revisions([rev1, rev3, b'foobar:']))

    def test_NULL(self):
        # NULL_REVISION is always present. So for
        # compatibility with 'has_revision' we make this work.
        repo = self.make_repository('.')
        self.assertEqual({NULL_REVISION},
                         repo.has_revisions([NULL_REVISION]))



# Copyright (C) 2009-2018 Jelmer Vernooij <jelmer@jelmer.uk>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Custom revision specifier for Subversion."""

# Please note that imports are delayed as much as possible here since
# if DWIM revspecs are supported this module is imported by __init__.py.

from ..errors import InvalidRevisionId
from ..revision import NULL_REVISION
from ..revisionspec import InvalidRevisionSpec, RevisionInfo, RevisionSpec


def valid_git_sha1(hex):
    """Check if `hex` is a validly formatted Git SHA1.

    :param hex: Hex string to validate
    :return: Boolean
    """
    try:
        int(hex, 16)
    except ValueError:
        return False
    else:
        return True


class RevisionSpec_git(RevisionSpec):
    """Selects a revision using a Git commit SHA1."""

    help_txt = """Selects a revision using a Git commit SHA1.

    Selects a revision using a Git commit SHA1, short or long.

    This works for both native Git repositories and Git revisions
    imported into Bazaar repositories.
    """

    prefix = 'git:'
    wants_revision_history = False

    def _lookup_git_sha1(self, branch, sha1):
        from .errors import GitSmartRemoteNotSupported
        from .mapping import default_mapping

        bzr_revid = getattr(branch.repository, "lookup_foreign_revision_id",
                            default_mapping.revision_id_foreign_to_bzr)(sha1)
        try:
            if branch.repository.has_revision(bzr_revid):
                return bzr_revid
        except GitSmartRemoteNotSupported:
            return bzr_revid
        raise InvalidRevisionSpec(self.user_spec, branch)

    def __nonzero__(self):
        # The default implementation uses branch.repository.has_revision()
        if self.rev_id is None:
            return False
        if self.rev_id == NULL_REVISION:
            return False
        return True

    def _find_short_git_sha1(self, branch, sha1):
        from .mapping import ForeignGit, mapping_registry
        parse_revid = getattr(branch.repository, "lookup_bzr_revision_id",
                              mapping_registry.parse_revision_id)
        def matches_revid(revid):
            if revid == NULL_REVISION:
                return False
            try:
                foreign_revid, mapping = parse_revid(revid)
            except InvalidRevisionId:
                return False
            if not isinstance(mapping.vcs, ForeignGit):
                return False
            return foreign_revid.startswith(sha1)
        with branch.repository.lock_read():
            graph = branch.repository.get_graph()
            last_revid = branch.last_revision()
            if matches_revid(last_revid):
                return last_revid
            for revid, _ in graph.iter_ancestry([last_revid]):
                if matches_revid(revid):
                    return revid
            raise InvalidRevisionSpec(self.user_spec, branch)

    def _as_revision_id(self, context_branch):
        loc = self.spec.find(':')
        git_sha1 = self.spec[loc + 1:].encode("utf-8")
        if (len(git_sha1) > 40 or len(git_sha1) < 4 or
                not valid_git_sha1(git_sha1)):
            raise InvalidRevisionSpec(self.user_spec, context_branch)
        from . import lazy_check_versions
        lazy_check_versions()
        if len(git_sha1) == 40:
            return self._lookup_git_sha1(context_branch, git_sha1)
        else:
            return self._find_short_git_sha1(context_branch, git_sha1)

    def _match_on(self, branch, revs):
        revid = self._as_revision_id(branch)
        return RevisionInfo.from_revision_id(branch, revid)

    def needs_branch(self):
        return True

    def get_branch(self):
        return None
